"""
This module contains classes and functions to handle files and directories.
"""
from __future__ import annotations
from datetime import datetime

import contextlib
import os
import io
import shutil
import tempfile
from typing import List, Union, Optional
import shutil
import zipfile
from pathlib import Path
# from System.IO import FileInfo  # public FileInfo (string fileName);

from SIMULTAN.Data.Assets import ResourceEntry, ResourceFileEntry, ContainedResourceFileEntry, Asset, ResourceDirectoryEntry
from SIMULTAN.Data.Taxonomy import SimTaxonomyEntry, SimTaxonomyEntryReference, SimTaxonomy
from SIMULTAN.Data.Components import SimComponent, ComponentMapping

from System.IO import DirectoryInfo as SystemDirectoryInfo
from System.IO import FileInfo as SystemFileInfo

# from .config import default_data_model

from . import config, logger

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .data_model import DataModel
    from .simultan_object import SimultanObject


@contextlib.contextmanager
def cd(newdir, cleanup=lambda: True):
    prevdir = os.getcwd()
    os.chdir(os.path.expanduser(newdir))
    try:
        yield
    finally:
        os.chdir(prevdir)
        cleanup()


@contextlib.contextmanager
def tempdir():
    dir_path = tempfile.mkdtemp()

    def cleanup():
        shutil.rmtree(dir_path)

    with cd(dir_path, cleanup):
        yield dir_path


def add_tag_to_resource(resource: Union[ResourceFileEntry, ContainedResourceFileEntry, ResourceDirectoryEntry],
                        tag: Union[SimTaxonomyEntry, SimTaxonomyEntryReference]):
    """
    Add a tag to an asset.

    :param resource: The resource to add the tag to.
    :param tag: The tag to add to the asset.
    :return: None
    """
    if isinstance(tag, SimTaxonomyEntry):
        tag = SimTaxonomyEntryReference(tag)

    if tag not in resource.Tags:
        resource.Tags.Add(tag)


def add_asset_to_component(comp: [SimComponent, SimultanObject],
                           asset: Union[ResourceFileEntry, ContainedResourceFileEntry, ResourceDirectoryEntry],
                           content_id: str = '',
                           tag: SimTaxonomyEntry = None) -> Asset:
    """
    Add an asset to a component with a content id.
    :param comp: Component to add the asset; ParameterStructure.Components.SimComponent
    :param asset: Asset to be added; ParameterStructure.Assets.ResourceFileEntry
    :param content_id: Content id of the asset; string; E.g. '0' as the page of a pdf
    :param tag: Tag to be added to the asset.
    :return:
    """
    wrapped_obj = comp if isinstance(comp, SimComponent) else comp._wrapped_obj

    if tag is not None:
        try:
            add_tag_to_resource(asset, tag)
        except Exception as e:
            logger.error(f'Error adding tag to asset {asset}: {e} ')
            raise e

    try:
        return ComponentMapping.AddAsset(wrapped_obj, asset, content_id)
    except Exception as e:
        logger.error(f'Error adding asset {asset} to component: {e}')
        raise e


def remove_asset_from_component(comp: Union[SimComponent, SimultanObject],
                                asset: Asset) -> None:
    """
    Remove an asset from a component with a content id.
    :param comp: Component to remove the asset from; ParameterStructure.Components.SimComponent
    :param asset:
    :return:
    """
    wrapped_obj = comp if isinstance(comp, SimComponent) else comp._wrapped_obj
    return ComponentMapping.RemoveAsset(wrapped_obj, asset)


def create_asset_from_string(filename: str,
                             content: str,
                             data_model: DataModel,
                             target_dir: Optional[Union[DirectoryInfo, ResourceDirectoryEntry, str]] = None,
                             tag: Optional[Union[SimTaxonomyEntry, SimTaxonomyEntryReference]] = None) -> ResourceFileEntry:
    """
    Create a new asset from a string. The asset is added to the data model.
    :param filename: Name of the file to be created. E.g. 'new_file.txt'
    :param content:  Content of the file. E.g. 'This is the content of the file.'
    :param data_model:  Data model to add the asset to.
    :param target_dir: Target directory to add the asset to.
    :param tag: Tag to be added to the asset.
    :return: ResourceFileEntry
    """
    with tempdir() as dirpath:
        filepath = os.path.join(dirpath, filename)
        with open(filepath, 'w') as f:
            f.write(content)

        if target_dir is not None:
            if isinstance(target_dir, DirectoryInfo):
                target_dir = target_dir.full_path

            resource = data_model.add_resource(filepath,
                                               target_dir=target_dir)
        else:
            resource = data_model.add_resource(filepath)

    if tag is not None:
        add_tag_to_resource(resource,
                            tag)

    return resource


def create_asset_from_str_io(filename: str,
                             content: io.StringIO,
                             data_model: DataModel,
                             target_dir: Optional[Union[DirectoryInfo, ResourceDirectoryEntry, str]] = None,
                             tag: Union[SimTaxonomyEntry, SimTaxonomyEntryReference] = None) -> ResourceFileEntry:
    """
    Create a new asset from a string io. The asset is added to the data model.
    :param filename: Name of the file to be created. E.g. 'new_file.txt'
    :param content:  Content of the file. E.g. 'This is the content of the file.'
    :param data_model:  Data model to add the asset to.
    :param target_dir: Target directory to add the asset to.
    :param tag: Tag to be added to the asset.
    :return: ResourceFileEntry
    """
    with tempdir() as dirpath:
        filepath = os.path.join(dirpath, filename)
        with open(filepath, 'w') as f:
            f.write(content.getvalue())

        resource = data_model.add_resource(filepath,
                                           target_dir=target_dir)

    if tag is not None:
        add_tag_to_resource(resource, tag)

    return resource


def create_asset_from_file(file_info: FileInfo,
                           data_model: DataModel,
                           tag: Union[SimTaxonomyEntry, SimTaxonomyEntryReference] = None) -> Union[
                           ResourceFileEntry, ContainedResourceFileEntry]:
    """
    Create a new asset from a file. The asset is added to the data model.
    :param file_info: FileInfo object of the file to be added.
    :param data_model:  Data model to add the asset to.
    :param tag: Tag to be added to the asset.
    :return: ResourceFileEntry
    """
    resource = data_model.add_resource(file_info.file_path)

    if tag is not None:
        add_tag_to_resource(resource, tag)

    return resource


def add_directory(data_model: DataModel,
                  directory: str,
                  parent_directory: Optional[Union[DirectoryInfo, ResourceDirectoryEntry, str]] = None,
                  tag: Union[SimTaxonomyEntry, SimTaxonomyEntryReference] = None) -> ResourceDirectoryEntry:

    """
    Add a directory to the data model.
    :param data_model:
    :param target_dir:
    :param tag:
    :return:
    """

    # create the directory
    resource_directory_entry = data_model.create_resource_directory(parent_directory=parent_directory)

    for filename in os.listdir(directory):
        file_path = os.path.join(directory, filename)
        resource = data_model.add_resource(file_path)
        if tag is not None:
            add_tag_to_resource(resource, tag)





class MetaMock(type):
    def __call__(cls, *args, **kwargs):
        resource_entry = kwargs.get('resource_entry', None)
        if resource_entry is not None and hasattr(resource_entry, 'Key'):
            obj = cls._cls_instances.get(resource_entry.Key, None)
            if obj is not None:
                return obj

        obj = cls.__new__(cls)
        obj.__init__(*args, **kwargs)
        if obj.resource_entry is not None:
            cls._cls_instances[obj.resource_entry.Key] = obj
        return obj


class DirectoryInfoMetaMock(type):

    def __call__(cls, *args, **kwargs):
        resource_entry: Optional[ResourceDirectoryEntry] = kwargs.get('resource_entry', None)
        if resource_entry is not None and hasattr(resource_entry, 'Key'):
            obj = cls._cls_instances.get(resource_entry.Key, None)
            if obj is not None:
                return obj

        obj = cls.__new__(cls)
        obj.__init__(*args, **kwargs)
        if obj.resource_entry is not None:
            cls._cls_instances[obj.resource_entry.Key] = obj
        return obj


class FileInfo(object, metaclass=MetaMock):

    _cls_instances = {}

    @classmethod
    def from_string(cls,
                    filename: str,
                    content: str,
                    target_dir: Optional[Union[DirectoryInfo, ResourceDirectoryEntry, str]] = None,
                    *args,
                    **kwargs,
                    ) -> FileInfo:
        """
        Create a file info object from a string.
        :param filename: Name of the file to be created. E.g. 'new_file.txt'
        :param content:  Content of the file. E.g. 'This is the content of the file.'
        :param target_dir: Target directory to add the asset to.
        :param args:
        :param kwargs:
        :return: FileInfo
        """

        data_model = kwargs.get('data_model', config.get_default_data_model())
        resource = create_asset_from_string(filename,
                                            content,
                                            target_dir=target_dir,
                                            *args,
                                            **kwargs)

        file_info = cls(resource_entry=resource,
                        data_model=data_model)
        file_info.write_content(content)
        return file_info

    @classmethod
    def from_existing_file(cls,
                           file_path: str,
                           *args,
                           **kwargs) -> FileInfo:

        data_model = kwargs.get('data_model', config.get_default_data_model())
        resource = data_model.add_resource_file(file_path)

        return cls(resource_entry=resource,
                   *args,
                   **kwargs)


    def __init__(self, file_path=None, *args, **kwargs):
        """
        Custom file info object to be used with the with statement. This object is used to open a file and close it
        automatically.
        Example:

        file_path = 'path/to/file.txt'
        file_info = FileInfo(file_path, 'r')

        with file_info as f:
            print(f.read())

        :param file_path:
        :param args:
        :param kwargs:
        """
        # do custom stuff here
        self._resource_entry: Union[ResourceFileEntry, ContainedResourceFileEntry, None] = None

        if file_path is not None:
            self.file_path: str = file_path
        else:
            self.file_path = kwargs.get('resource_entry').File.FullPath

        self.data_model: Union[DataModel, None] = kwargs.get('data_model', None)
        self.resource_entry = kwargs.get('resource_entry', None)

        self.encoding = kwargs.get('encoding', 'utf-8')

        self.args = args
        self.kwargs = kwargs

    @property
    def parent(self):
        return self.resource_entry.Parent

    @property
    def key(self) -> int:
        try:
            return self.resource_entry.Key
        except Exception as e:
            return None

    @property
    def directory(self) -> DirectoryInfo:
        return DirectoryInfo(resource_entry=self.resource_entry.Parent,
                             data_model=self.data_model)

    @property
    def resource_entry(self) -> Union[ResourceFileEntry, ContainedResourceFileEntry, None]:
        if self._resource_entry is None:
            if self.data_model is None:
                logger.warning(f'No data model provided. Using default data model: {config.get_default_data_model().id}.')
                self.data_model = config.get_default_data_model()
            if self.data_model is not None:
                self.resource_entry = self.data_model.add_resource(self.file_path)
                self.file_path = self.resource_entry.File.FullPath
        return self._resource_entry

    @resource_entry.setter
    def resource_entry(self, value):

        if value is not None:
            self._cls_instances[value.Key] = self
        else:
            del self._cls_instances[self._resource_entry.Key]
        self._resource_entry = value

    @property
    def file_size(self) -> Optional[int]:
        try:
            return os.path.getsize(self.file_path)
        except FileNotFoundError:
            raise FileNotFoundError(f'File not found: {self.file_path}')
        except Exception as e:
            raise e

    @property
    def last_modified(self) -> datetime:
        return datetime.fromtimestamp(os.path.getmtime(self.file_path))

    @resource_entry.setter
    def resource_entry(self, value):
        self._resource_entry = value

    @property
    def filename(self) -> str:
        return self.resource_entry.File.Name

    @property
    def name(self) -> str:
        return os.path.basename(self.file_path)

    @name.setter
    def name(self, value: str):
        os.rename(self.file_path, os.path.join(os.path.dirname(self.file_path), value))
        self.file_path = os.path.join(os.path.dirname(self.file_path), value)

    @property
    def full_path(self) -> str:
        return os.path.abspath(self.file_path)

    @property
    def size(self) -> int:
        return os.path.getsize(self.file_path)

    @property
    def is_zip(self) -> bool:
        return self.file_path.endswith('.zip')

    @property
    def files(self) -> List[str]:
        if self.is_zip:
            with zipfile.ZipFile(self.file_path, 'r') as z:
                return z.namelist()
        else:
            return []

    @property
    def exists(self) -> bool:
        return os.path.exists(self.file_path)

    @property
    def content(self) -> str:
        return self.get_content(encoding=self.encoding)

    @content.setter
    def content(self, value: str):
        self.write_content(value)

    def __enter__(self):
        self.file_obj = open(self.file_path, *self.args, **self.kwargs)
        return self.file_obj

    def __exit__(self, *args):
        self.file_obj.close()

    def __repr__(self):
        return f'FileInfo({self.file_path})'

    def move(self, new_directory_path: Union[str, DirectoryInfo]) -> FileInfo:

        if isinstance(new_directory_path, str):
            new_directory_path = DirectoryInfo.from_existing_directory(new_directory_path,
                                                                       data_model=self.data_model)

        # check if file can be moved
        try:
            self.resource_entry.ChangeLocation(SystemDirectoryInfo(new_directory_path.full_path),
                                                   '1',
                                                   True)
        except Exception as e:
            logger.error(f'Error moving file: {e}')
            raise e

    def get_content(self, encoding='utf-8') -> Optional[Union[str, dict[str, str]]]:
        """
        Get the content of the file.
        :param encoding: Encoding of the file.
        :return: File content
        """
        if self.exists:
            if self.file_path.endswith('.zip'):
                content = {}
                with zipfile.ZipFile(self.file_path, 'r') as z:
                    for file in z.namelist():
                        content[file] = z.read(file).decode(encoding)
                return content
            else:
                with open(self.file_path, 'r', encoding=encoding) as f:
                    return f.read()
        else:
            return

    def copy(self,
             new_file_path: Union[str, DirectoryInfo]) -> FileInfo:
        """
        Copy the file to a new location.
        :param new_file_path: New file path.
        :return: FileInfo
        """
        if isinstance(new_file_path, DirectoryInfo):
            new_file_path = os.path.join(new_file_path.full_path, self.filename)

        shutil.copy(self.full_path, new_file_path)
        return FileInfo.from_existing_file(file_path=new_file_path,
                                           data_model=self.data_model)

    def write_content(self, content: str, encoding='utf-8') -> None:
        """
        Write content to the file.
        :param content: Content to be written to the file.
        :param encoding: Encoding of the file.
        :return: None
        """
        with open(self.file_path, 'w', encoding=encoding) as f:
            f.write(content)

    def append_content(self, content: str, encoding='utf-8') -> None:
        """
        Append content to the file.
        :param content: Content to be appended to the file.
        :param encoding: Encoding of the file.
        :return: None
        """
        with open(self.file_path, 'a', encoding=encoding) as f:
            f.write(content)

    def delete(self) -> None:
        """
        Delete the file.
        :return: None
        """
        if self.resource_entry is not None:
            if self.resource_entry.Key in self._cls_instances:
                del self._cls_instances[self.resource_entry.Key]
            self.data_model.delete_resource(self.resource_entry)

        os.remove(self.file_path)

    def to_json(self) -> dict:

        obj_dict = {
            'key': self.key,
            'name': self.name,
            'file_path': self.file_path,
            'file_size': self.file_size,
            'last_modified': self.last_modified,
            'encoding': self.encoding,
            'is_zip': self.is_zip,
        }

        return obj_dict

    def json_ref(self):
        return {"$ref": {
            "$type": 'FileInfo',
            "$key": str(self.key)
        }
        }


class DirectoryInfo(object, metaclass=DirectoryInfoMetaMock):

    _cls_instances = {}

    @classmethod
    def get_by_key(cls, key: int) -> Optional[DirectoryInfo]:
        return cls._cls_instances.get(key, None)

    @classmethod
    def from_existing_directory(cls,
                                directory_path: str,
                                add_files: bool = True,
                                add_sub_directories: bool = True,
                                *args,
                                **kwargs) -> DirectoryInfo:

        data_model = kwargs.get('data_model', config.get_default_data_model())

        if not directory_path.startswith(str(data_model.project.ProjectUnpackFolder)):
            directory_path = os.path.join(str(data_model.project.ProjectUnpackFolder), directory_path)

        res = data_model.project_data_manager.AssetManager.CreateResourceDirIn(os.path.basename(directory_path),
                                                     SystemDirectoryInfo(os.path.dirname(directory_path)),
                                                     '')

        resource = data_model.project_data_manager.AssetManager.GetResource(res.Item1)

        directory_info = cls(resource_entry=resource,
                             *args,
                             **kwargs)


        if add_files:
            directory_info.add_all_contained_files()

            # for file in os.listdir(directory_path):
            #     file_path = os.path.join(directory_path, file)
            #     data_model.add_resource_file(file_path, target_dir=resource)

        if add_sub_directories:
            directory_info.add_all_contained_directories()

            # for sub_dir in os.listdir(directory_path):
            #     sub_dir_path = os.path.join(directory_path, sub_dir)
            #     data_model.add_resource_directory(sub_dir_path, parent_directory=resource)

        return directory_info

    def __init__(self,
                 path: Optional[str] = None,
                 helper_file: Optional[FileInfo] = None,
                 resource_entry: Optional[ResourceDirectoryEntry] = None,
                 *args,
                 **kwargs):

        self._resource_entry: Optional[ResourceDirectoryEntry] = None
        self._helper_file: Optional[FileInfo] = None
        self.data_model: Optional[DataModel] = kwargs.get('data_model', None)
        self.path: str = path

        self.resource_entry = resource_entry
        self.helper_file = helper_file

    @property
    def tags(self) -> List[SimTaxonomyEntry]:
        return list(self.resource_entry.Tags)

    @property
    def full_path(self) -> str:
        return self.resource_entry.CurrentFullPath

    @property
    def relative_path(self) -> str:
        return self.resource_entry.CurrentRelativePath

    @property
    def helper_file(self) -> Optional[FileInfo]:
        if self._helper_file is None or not isinstance(self._helper_file, FileInfo):
            self._helper_file = self.add_file('__dir_helper_file__')

        return self._helper_file

    @helper_file.setter
    def helper_file(self, value):
        self._helper_file = value

    @property
    def resource_entry(self) -> Optional[ResourceDirectoryEntry]:
        if self._resource_entry is None:
            if self.data_model is None:
                logger.warning(
                    f'No data model provided. Using default data model: {config.get_default_data_model().id}.')
                self.data_model = config.get_default_data_model()
            if self.data_model is not None:
                self.resource_entry = self.data_model.create_resource_directory(self.path)
                self._cls_instances[self.resource_entry.Key] = self
                self.path = self.resource_entry.CurrentFullPath
        return self._resource_entry

    @resource_entry.setter
    def resource_entry(self, value):

        orig_value = self._resource_entry
        self._resource_entry = value

        if self._resource_entry is None:
            if orig_value is not None:
                del self._cls_instances[orig_value.Key]
            return

        if self.key is not None:
            if value is not None:
                self._cls_instances[value.Key] = self
            else:
                del self._cls_instances[self._resource_entry.Key]
            self._resource_entry = value

    @property
    def parent(self) -> Optional[ResourceDirectoryEntry]:
        if self.resource_entry.Parent is not None:
            if self.resource_entry.Parent.Key in self._cls_instances:
                return self.get_by_key(self.resource_entry.Parent.Key)
            return DirectoryInfo(resource_entry=self.resource_entry.Parent)
        else:
            return self.resource_entry.Parent

    @property
    def sub_directories(self) -> List[DirectoryInfo]:
        return [DirectoryInfo(resource_entry=entry,
                              data_model=self.data_model) for entry in self.resource_entry.Children if isinstance(entry, ResourceDirectoryEntry)]

    @property
    def files(self) -> List[FileInfo]:
        return [FileInfo(resource_entry=entry,
                         data_model=self.data_model) for entry in self.resource_entry.Children if isinstance(entry,
                                                                                                             (
                                                                                                             ResourceFileEntry,
                                                                                                             ContainedResourceFileEntry)
                                                                                                             ) and entry.Name != '__dir_helper_file__'
                ]

    @property
    def key(self) -> Optional[int]:
        if self.resource_entry is not None:
            return self.resource_entry.Key
        else:
            return None

    def add_sub_directory(self, dirname: str) -> DirectoryInfo:

        existing = next((x for x in self.sub_directories if x.resource_entry.current_relative_path == dirname), None)
        if existing is not None:
            return existing

        return DirectoryInfo(path=os.path.join(self.resource_entry.current_relative_path, dirname),
                             data_model=self.data_model)

    def get_sub_directory(self,
                          dirname: str,
                          create=False) -> Optional[DirectoryInfo]:
        """
        Get a sub directory by name.
        :param dirname: directory name, relative to self.relative_path
        :param create:
        :return:
        """

        logger.debug(f'Getting sub directory in {self.relative_path}: {dirname}')
        res = next((x for x in self.sub_directories if x.resource_entry.Name == dirname), None)
        if not res and create:
            logger.debug(f'Creating sub directory in {self.relative_path}: {dirname}')
            res = self.add_sub_directory(dirname)

        return res

    def file_exists(self, filename: str) -> bool:
        return any(x for x in self.files if x.resource_entry.Name == filename)

    def get_file(self, filename: str) -> Optional[FileInfo]:
        resource = next((x for x in self.files if x.resource_entry.Name == filename), None)
        if resource is not None:
            return FileInfo(resource_entry=resource,
                            data_model=self.data_model)
        else:
            return None

    def add_file(self,
                 filename: str,
                 content: Optional[str] = None) -> FileInfo:

        if Path(self.full_path).joinpath(filename).exists():
            return FileInfo.from_existing_file(str(Path(self.full_path).joinpath(filename)), data_model=self.data_model)


        if self.resource_entry.Children:
            file = self.get_file(filename)
            if file is not None:
                return file

        if content is not None:
            return FileInfo.from_string(filename=filename,
                                        content=content,
                                        target_dir=self.resource_entry,
                                        data_model=self.data_model)
        else:
            new_resource = self.data_model.add_empty_resource(filename=os.path.join(self.full_path, filename))
            return FileInfo(resource_entry=new_resource,
                            data_model=self.data_model)

    def add_all_contained_files(self):
        for file in os.listdir(self.full_path):
            full_filename = os.path.join(self.full_path, file)
            if Path(full_filename).is_file():

                if full_filename in (x.current_full_path for x in self.resource_entry.Children):
                    continue
                else:
                    logger.info(f'Adding file: {full_filename} to resources')
                    FileInfo.from_existing_file(full_filename, data_model=self.data_model)

    def add_all_contained_directories(self):
        for file in os.listdir(self.full_path):
            full_filename = os.path.join(self.full_path, file)
            if Path(full_filename).is_dir():
                if full_filename in (x.current_full_path for x in self.resource_entry.Children):
                    continue
                else:
                    logger.info(f'Adding directory: {full_filename} to resources')
                    DirectoryInfo.from_existing_directory(full_filename, data_model=self.data_model)

        for directory in self.sub_directories:
            directory.add_all_contained_files()
            directory.add_all_contained_directories()

    def add_tag(self, tag: SimTaxonomyEntry) -> None:
        add_tag_to_resource(self.resource_entry, tag)

    def __repr__(self):
        return f'DirectoryInfo(key:{self.key}, hash: {hash(self)}; {self.full_path})'



import typing, abc
from System import Exception, Guid, DateTime, IComparable, IComparable_1, ICloneable, IEquatable_1
from System.Collections import IDictionary
from System.Reflection import MethodBase
from System.Collections.ObjectModel import ObservableCollection_1, ReadOnlyDictionary_2
from SIMULTAN.Data import SimId, SimObjectNew_1, SimManagedCollection_1, IReferenceLocation
from System.Globalization import CultureInfo
from System.Collections.Generic import IEnumerable_1, Dictionary_2
from SIMULTAN.Projects import ProjectData

class DeleteActionAlreadyRegisteredException(Exception):
    def __init__(self, message: str, innerException: Exception) -> None: ...
    @property
    def Data(self) -> IDictionary: ...
    @property
    def HelpLink(self) -> str: ...
    @HelpLink.setter
    def HelpLink(self, value: str) -> str: ...
    @property
    def HResult(self) -> int: ...
    @HResult.setter
    def HResult(self, value: int) -> int: ...
    @property
    def InnerException(self) -> Exception: ...
    @property
    def Message(self) -> str: ...
    @property
    def Source(self) -> str: ...
    @Source.setter
    def Source(self, value: str) -> str: ...
    @property
    def StackTrace(self) -> str: ...
    @property
    def TargetSite(self) -> MethodBase: ...


class ISimTaxonomyElement(typing.Protocol):
    @property
    def Taxonomy(self) -> SimTaxonomy: ...
    @abc.abstractmethod
    def NotifyLocalizationChanged(self) -> None: ...


class SimChildTaxonomyEntryCollection(ObservableCollection_1[SimTaxonomyEntry]):
    def __init__(self, owner: SimTaxonomyEntry) -> None: ...
    @property
    def Count(self) -> int: ...
    @property
    def Item(self) -> SimTaxonomyEntry: ...
    @Item.setter
    def Item(self, value: SimTaxonomyEntry) -> SimTaxonomyEntry: ...


class SimPlaceholderTaxonomyEntryReference(SimTaxonomyEntryReference):
    def __init__(self, placeholderName: str) -> None: ...
    @property
    def PlaceholderName(self) -> str: ...
    @property
    def Target(self) -> SimTaxonomyEntry: ...
    @property
    def TaxonomyEntryId(self) -> SimId: ...


class SimTaxonomy(SimObjectNew_1[SimTaxonomyCollection], ISimTaxonomyElement):
    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, id: SimId) -> None: ...
    @typing.overload
    def __init__(self, key: str, name: str, description: str, culture: CultureInfo = ...) -> None: ...
    @typing.overload
    def __init__(self, name: str, description: str = ..., culture: CultureInfo = ...) -> None: ...
    @property
    def Description(self) -> str: ...
    @Description.setter
    def Description(self, value: str) -> str: ...
    @property
    def Entries(self) -> SimTaxonomyEntryCollection: ...
    @property
    def Factory(self) -> SimTaxonomyCollection: ...
    @Factory.setter
    def Factory(self, value: SimTaxonomyCollection) -> SimTaxonomyCollection: ...
    @property
    def GlobalID(self) -> Guid: ...
    @property
    def Id(self) -> SimId: ...
    @Id.setter
    def Id(self, value: SimId) -> SimId: ...
    @property
    def IsDeletable(self) -> bool: ...
    @IsDeletable.setter
    def IsDeletable(self, value: bool) -> bool: ...
    @property
    def IsReadonly(self) -> bool: ...
    @IsReadonly.setter
    def IsReadonly(self, value: bool) -> bool: ...
    @property
    def Key(self) -> str: ...
    @Key.setter
    def Key(self, value: str) -> str: ...
    @property
    def Languages(self) -> SimTaxonomyLanguageCollection: ...
    @property
    def LocalID(self) -> int: ...
    @property
    def Localization(self) -> SimTaxonomyLocalization: ...
    @property
    def Taxonomy(self) -> SimTaxonomy: ...
    def GetAllEntriesFlat(self) -> IEnumerable_1[SimTaxonomyEntry]: ...
    def GetTaxonomyEntryByKey(self, key: str) -> SimTaxonomyEntry: ...
    def IsIdentical(self, other: SimTaxonomy) -> bool: ...
    def IsKeyInUse(self, key: str) -> bool: ...
    def NotifyLocalizationChanged(self) -> None: ...


class SimTaxonomyCollection(SimManagedCollection_1[SimTaxonomy]):
    def __init__(self, owner: ProjectData) -> None: ...
    @property
    def CalledFromLocation(self) -> IReferenceLocation: ...
    @property
    def Count(self) -> int: ...
    @property
    def HasChanges(self) -> bool: ...
    @property
    def IsClosing(self) -> bool: ...
    @IsClosing.setter
    def IsClosing(self, value: bool) -> bool: ...
    @property
    def IsLoading(self) -> bool: ...
    @IsLoading.setter
    def IsLoading(self, value: bool) -> bool: ...
    @property
    def IsMergeInProgress(self) -> bool: ...
    @IsMergeInProgress.setter
    def IsMergeInProgress(self, value: bool) -> bool: ...
    @property
    def Item(self) -> SimTaxonomy: ...
    @Item.setter
    def Item(self, value: SimTaxonomy) -> SimTaxonomy: ...
    @property
    def LastChange(self) -> DateTime: ...
    @property
    def ProjectData(self) -> ProjectData: ...
    def FindEntry(self, taxonomyKey: str, entryKey: str, isReadonly: bool = ...) -> SimTaxonomyEntry: ...
    def Merge(self, other: SimTaxonomyCollection) -> Dictionary_2[SimTaxonomy, SimTaxonomy]: ...
    def MergeWithDefaults(self, defaults: SimTaxonomyCollection) -> bool: ...
    def StartLoading(self) -> None: ...
    def StopLoading(self) -> None: ...
    # Skipped GetTaxonomyByKeyOrName due to it being static, abstract and generic.

    GetTaxonomyByKeyOrName : GetTaxonomyByKeyOrName_MethodGroup
    class GetTaxonomyByKeyOrName_MethodGroup:
        @typing.overload
        def __call__(self, key: str, name: str = ..., culture: CultureInfo = ...) -> SimTaxonomy:...
        @typing.overload
        def __call__(self, taxonomies: IEnumerable_1[SimTaxonomy], key: str, name: str = ..., culture: CultureInfo = ...) -> SimTaxonomy:...



class SimTaxonomyEntry(SimObjectNew_1[SimTaxonomyCollection], IComparable, IComparable_1[SimTaxonomyEntry], ISimTaxonomyElement):
    @typing.overload
    def __init__(self, id: SimId, key: str) -> None: ...
    @typing.overload
    def __init__(self, id: SimId, key: str, name: str, description: str = ..., culture: CultureInfo = ...) -> None: ...
    @typing.overload
    def __init__(self, key: str) -> None: ...
    @typing.overload
    def __init__(self, key: str, name: str, description: str = ..., culture: CultureInfo = ...) -> None: ...
    @property
    def Children(self) -> SimChildTaxonomyEntryCollection: ...
    @property
    def Description(self) -> str: ...
    @Description.setter
    def Description(self, value: str) -> str: ...
    @property
    def Factory(self) -> SimTaxonomyCollection: ...
    @Factory.setter
    def Factory(self, value: SimTaxonomyCollection) -> SimTaxonomyCollection: ...
    @property
    def GlobalID(self) -> Guid: ...
    @property
    def Id(self) -> SimId: ...
    @Id.setter
    def Id(self, value: SimId) -> SimId: ...
    @property
    def Key(self) -> str: ...
    @Key.setter
    def Key(self, value: str) -> str: ...
    @property
    def LocalID(self) -> int: ...
    @property
    def Localization(self) -> SimTaxonomyLocalization: ...
    @property
    def Parent(self) -> SimTaxonomyEntry: ...
    @Parent.setter
    def Parent(self, value: SimTaxonomyEntry) -> SimTaxonomyEntry: ...
    @property
    def Taxonomy(self) -> SimTaxonomy: ...
    @Taxonomy.setter
    def Taxonomy(self, value: SimTaxonomy) -> SimTaxonomy: ...
    def Matches(self, taxonomyKey: str, entryKey: str) -> bool: ...
    def NotifyLocalizationChanged(self) -> None: ...
    def OnIsBeingDeleted(self) -> None: ...
    # Skipped CompareTo due to it being static, abstract and generic.

    CompareTo : CompareTo_MethodGroup
    class CompareTo_MethodGroup:
        @typing.overload
        def __call__(self, other: SimTaxonomyEntry) -> int:...
        @typing.overload
        def __call__(self, obj: typing.Any) -> int:...



class SimTaxonomyEntryCollection(ObservableCollection_1[SimTaxonomyEntry]):
    def __init__(self, owner: SimTaxonomy) -> None: ...
    @property
    def Count(self) -> int: ...
    @property
    def Item(self) -> SimTaxonomyEntry: ...
    @Item.setter
    def Item(self, value: SimTaxonomyEntry) -> SimTaxonomyEntry: ...


class SimTaxonomyEntryOrString:
    @typing.overload
    def __init__(self, original: SimTaxonomyEntryOrString) -> None: ...
    @typing.overload
    def __init__(self, taxonomyEntry: SimTaxonomyEntry) -> None: ...
    @typing.overload
    def __init__(self, taxonomyEntry: SimTaxonomyEntryReference) -> None: ...
    @typing.overload
    def __init__(self, text: str) -> None: ...
    @property
    def HasTaxonomyEntry(self) -> bool: ...
    @property
    def HasTaxonomyEntryReference(self) -> bool: ...
    @property
    def TaxonomyEntryReference(self) -> SimTaxonomyEntryReference: ...
    @property
    def Text(self) -> str: ...
    @property
    def TextOrKey(self) -> str: ...
    def GetHashCode(self) -> int: ...
    # Skipped Equals due to it being static, abstract and generic.

    Equals : Equals_MethodGroup
    class Equals_MethodGroup:
        @typing.overload
        def __call__(self, other: SimTaxonomyEntryOrString) -> bool:...
        @typing.overload
        def __call__(self, obj: typing.Any) -> bool:...



class SimTaxonomyEntryReference(ICloneable, IEquatable_1[SimTaxonomyEntryReference], IComparable_1[SimTaxonomyEntryReference]):
    @typing.overload
    def __init__(self, other: SimTaxonomyEntryReference) -> None: ...
    @typing.overload
    def __init__(self, target: SimTaxonomyEntry) -> None: ...
    @property
    def Target(self) -> SimTaxonomyEntry: ...
    @property
    def TaxonomyEntryId(self) -> SimId: ...
    def Clone(self) -> typing.Any: ...
    def CompareTo(self, other: SimTaxonomyEntryReference) -> int: ...
    def GetHashCode(self) -> int: ...
    def Matches(self, taxonomyKey: str, entryKey: str) -> bool: ...
    def __eq__(self, a: SimTaxonomyEntryReference, b: SimTaxonomyEntryReference) -> bool: ...
    def __ne__(self, a: SimTaxonomyEntryReference, b: SimTaxonomyEntryReference) -> bool: ...
    # Skipped Equals due to it being static, abstract and generic.

    Equals : Equals_MethodGroup
    class Equals_MethodGroup:
        @typing.overload
        def __call__(self, other: SimTaxonomyEntryReference) -> bool:...
        @typing.overload
        def __call__(self, obj: typing.Any) -> bool:...



class SimTaxonomyEntryReferenceCollection(ObservableCollection_1[SimTaxonomyEntryReference]):
    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, collection: IEnumerable_1[SimTaxonomyEntryReference]) -> None: ...
    @property
    def Count(self) -> int: ...
    @property
    def Item(self) -> SimTaxonomyEntryReference: ...
    @Item.setter
    def Item(self, value: SimTaxonomyEntryReference) -> SimTaxonomyEntryReference: ...
    def RestoreDefaultTaxonomyReferences(self, projectData: ProjectData) -> None: ...


class SimTaxonomyLanguageCollection(ObservableCollection_1[CultureInfo]):
    def __init__(self, owner: SimTaxonomy) -> None: ...
    @property
    def Count(self) -> int: ...
    @property
    def Item(self) -> CultureInfo: ...
    @Item.setter
    def Item(self, value: CultureInfo) -> CultureInfo: ...


class SimTaxonomyLocalization:
    def __init__(self, owner: ISimTaxonomyElement) -> None: ...
    @property
    def Entries(self) -> ReadOnlyDictionary_2[CultureInfo, SimTaxonomyLocalizationEntry]: ...
    def IsIdenticalTo(self, other: SimTaxonomyLocalization) -> bool: ...
    def Localize(self, culture: CultureInfo = ...) -> SimTaxonomyLocalizationEntry: ...
    # Skipped SetLanguage due to it being static, abstract and generic.

    SetLanguage : SetLanguage_MethodGroup
    class SetLanguage_MethodGroup:
        @typing.overload
        def __call__(self, entry: SimTaxonomyLocalizationEntry) -> None:...
        @typing.overload
        def __call__(self, culture: CultureInfo, name: str, description: str) -> None:...



class SimTaxonomyLocalizationEntry:
    def __init__(self, culture: CultureInfo, name: str = ..., description: str = ...) -> None: ...
    @property
    def Culture(self) -> CultureInfo: ...
    @property
    def Description(self) -> str: ...
    @property
    def IsValid(self) -> bool: ...
    @property
    def Name(self) -> str: ...
    def Equals(self, obj: typing.Any) -> bool: ...
    def GetHashCode(self) -> int: ...
    def ToString(self) -> str: ...


class TaxonomyEntryNotFoundException(Exception):
    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, message: str) -> None: ...
    @typing.overload
    def __init__(self, message: str, innerException: Exception) -> None: ...
    @property
    def Data(self) -> IDictionary: ...
    @property
    def HelpLink(self) -> str: ...
    @HelpLink.setter
    def HelpLink(self, value: str) -> str: ...
    @property
    def HResult(self) -> int: ...
    @HResult.setter
    def HResult(self, value: int) -> int: ...
    @property
    def InnerException(self) -> Exception: ...
    @property
    def Message(self) -> str: ...
    @property
    def Source(self) -> str: ...
    @Source.setter
    def Source(self, value: str) -> str: ...
    @property
    def StackTrace(self) -> str: ...
    @property
    def TargetSite(self) -> MethodBase: ...




import resources
from src.PySimultan2.data_model import DataModel
from src.PySimultan2.object_mapper import PythonMapper
from src.PySimultan2.taxonomy_maps import TaxonomyMap, Content

try:
    import importlib.resources as pkg_resources
except ImportError:
    # Try backported to PY<37 importlib_resources.
    import importlib_resources as pkg_resources


with pkg_resources.path(resources, 'test_set_attr_property.simultan') as r_path:
    project_path = str(r_path)

data_model = DataModel.create_new_project(project_path=project_path,
                                          user_name='admin',
                                          password='admin')

mapper = PythonMapper()


class TestComponent(object):
    def __init__(self, *args, **kwargs):
        self.value = kwargs.get('value')
        self.value2 = kwargs.get('value2')


content0 = Content(text_or_key='value',  # text or key of the content/parameter/property
                   property_name='value',  # name of the generated property
                   type=None,  # type of the content/parameter/property
                   unit=None,  # unit of the content/parameter/property
                   documentation='value to test',
                   component_policy='subcomponent')

content1 = Content(text_or_key='value2',
                   property_name='value2',
                   type=None,
                   unit=None,
                   documentation='value2 to test',
                   component_policy='subcomponent',
                   ValueMin=100,
                   ValueMax=200)

test_component_map = TaxonomyMap(taxonomy_name='PySimultan',
                                 taxonomy_key='PySimultan',
                                 taxonomy_entry_name='TestComponent',
                                 taxonomy_entry_key='TestComponent',
                                 content=[content0, content1],
                                 )

mapper.register(test_component_map.taxonomy_entry_key, TestComponent, taxonomy_map=test_component_map)
cls1 = mapper.get_mapped_class(test_component_map.taxonomy_entry_key)

component = cls1(name='test_component',
                 value=78,
                 value2=350)

component.get_raw_attr('value')

component.set_attr_prop('value', 'ValueMin', 100)
component.set_attr_prop('value', 'ValueMax', 200)

data_model.save()
data_model.cleanup()
mapper.clear()


data_model = DataModel(project_path=project_path,
                       user_name='admin',
                       password='admin')


typed_data = mapper.get_typed_data(component_list=data_model.data, data_model=data_model)

pass



import os
from PySimultan2.src.PySimultan2 import DataModel, TaxonomyMap, Content, PythonMapper
from PySimultan2.src.PySimultan2.files import FileInfo, DirectoryInfo


project_dir = os.environ.get('PROJECT_DIR', '/simultan_projects')
if not os.path.exists(project_dir):
    os.makedirs(project_dir)


def test_files_and_directory_creation():
    new_data_model = DataModel.create_new_project(project_path=os.path.join(project_dir, 'test_dir_files.simultan'),
                                                  user_name='admin',
                                                  password='admin')

    print(new_data_model.file_directories)


    new_directory_info = DirectoryInfo(path='directory_info_dir',
                                       data_model=new_data_model)

    print(new_data_model.file_directories)

    print(new_directory_info.resource_entry)
    print(new_directory_info.parent)

    new_file1 = new_directory_info.add_file('test_add_file.txt', 'This is a test file')
    new_file2 = new_directory_info.add_file('test_add_file2.txt')

    sub_directory_info = new_directory_info.add_sub_directory('sub_dir')
    print(sub_directory_info.parent)

    print(new_directory_info.sub_directories)


    new_directory = new_data_model.create_resource_directory('test_dir')
    new_directory2 = new_data_model.create_resource_directory('test_dir2')
    new_directory4 = new_data_model.create_resource_directory('test_dir3')

    new_data_model.add_empty_resource(filename=os.path.join(new_directory.CurrentFullPath, 'test_empty_file.txt'))

    new_data_model.add_empty_resource(filename='test_empty_file2.txt',
                                      target_dir=new_directory)

    new_file_info0 = FileInfo.from_string(filename='test_file.txt',
                                          content='This is a test file',
                                          data_model=new_data_model)

    new_file_info = FileInfo.from_string(filename='test_file.txt',
                                         content='This is a test file',
                                         target_dir=new_directory,
                                         data_model=new_data_model)

    new_file_info2 = FileInfo.from_string(filename='test_file2.txt',
                                         content='This is a test file 2',
                                         target_dir=new_directory2.current_full_path,
                                         data_model=new_data_model)

    # create just a file in the directory
    with open(os.path.join(new_directory2.current_full_path, 'not_contained_test_file3.txt'), 'w') as f:
        f.write('This is a test file 3')


    new_data_model.save()
    new_data_model.cleanup()


def create_mapper():
    mapper = PythonMapper()

    class TestComponent(object):
        def __init__(self, *args, **kwargs):
            self.file = kwargs.get('file')
            self.directory = kwargs.get('directory')
            self.param_1 = kwargs.get('param_1')
            self.param_2 = kwargs.get('param_2')

    content0 = Content(text_or_key='file',  # text or key of the content/parameter/property
                       property_name='file',  # name of the generated property
                       type='file',  # type of the content/parameter/property
                       unit=None,  # unit of the content/parameter/property
                       documentation='file_value to test')

    content1 = Content(text_or_key='directory',  # text or key of the content/parameter/property
                       property_name='directory',  # name of the generated property
                       type='file',  # type of the content/parameter/property
                       unit=None,  # unit of the content/parameter/property
                       documentation='file_value_2 to test')

    content2 = Content(text_or_key='param_1',  # text or key of the content/parameter/property
                       property_name='param_1',  # name of the generated property
                       type=None,  # type of the content/parameter/property
                       unit=None,  # unit of the content/parameter/property
                       documentation='param_1 to test')

    content3 = Content(text_or_key='param_2',  # text or key of the content/parameter/property
                       property_name='param_2',  # name of the generated property
                       type=None,  # type of the content/parameter/property
                       unit=None,  # unit of the content/parameter/property
                       documentation='param_2 to test')

    test_component_map = TaxonomyMap(taxonomy_name='PySimultan',
                                     taxonomy_key='Test',
                                     taxonomy_entry_name='test_component',
                                     taxonomy_entry_key='test_component',
                                     content=[content0, content1, content2, content3],
                                     )

    mapper.register(test_component_map.taxonomy_entry_key, TestComponent, taxonomy_map=test_component_map)

    return mapper


def test_component_with_directory():

    new_data_model = DataModel.create_new_project(project_path=os.path.join(project_dir, 'test_dir_component.simultan'),
                                                  user_name='admin',
                                                  password='admin')

    mapper = create_mapper()
    mapped_test_component_cls = mapper.get_mapped_class('test_component')

    new_component = mapped_test_component_cls(file=FileInfo.from_string(filename='test_file.txt',
                                                                                    content='This is a test file',
                                                                                    data_model=new_data_model),
                                                directory=DirectoryInfo(path='test_dir',
                                                                        data_model=new_data_model),
                                                param_1='param_1_value',
                                                param_2=2)

    print(new_component.directory)

    new_data_model.save()
    new_data_model.cleanup()
    mapper.clear()


def test_load_component_with_directory():
    new_data_model = DataModel(project_path=os.path.join(project_dir, 'test_dir_component.simultan'))

    mapper = create_mapper()
    typed_data = new_data_model.get_typed_data(mapper=mapper)
    mapped_test_component_cls = mapper.get_mapped_class('test_component')

    assert isinstance(mapped_test_component_cls.cls_instances[0].directory,  DirectoryInfo)

    new_data_model.cleanup()
    mapper.clear()


if __name__ == '__main__':
    # test_files_and_directory_creation()
    test_component_with_directory()
    test_load_component_with_directory()



import typing, clr, abc
from System import Attribute, Array, Array_1, IDisposable, SystemException, Exception
from System.Collections import ICollection, IEnumerator, Hashtable, ArrayList, IDictionary
from System.Runtime.Serialization import IDeserializationCallback, SerializationInfo, StreamingContext
from System.Security.Permissions import PermissionState
from System.Reflection import AssemblyName, MethodInfo, MethodBase

class AllowPartiallyTrustedCallersAttribute(Attribute):
    def __init__(self) -> None: ...
    @property
    def PartialTrustVisibilityLevel(self) -> PartialTrustVisibilityLevel: ...
    @PartialTrustVisibilityLevel.setter
    def PartialTrustVisibilityLevel(self, value: PartialTrustVisibilityLevel) -> PartialTrustVisibilityLevel: ...
    @property
    def TypeId(self) -> typing.Any: ...


class IPermission(ISecurityEncodable, typing.Protocol):
    @abc.abstractmethod
    def Copy(self) -> IPermission: ...
    @abc.abstractmethod
    def Demand(self) -> None: ...
    @abc.abstractmethod
    def Intersect(self, target: IPermission) -> IPermission: ...
    @abc.abstractmethod
    def IsSubsetOf(self, target: IPermission) -> bool: ...
    @abc.abstractmethod
    def Union(self, target: IPermission) -> IPermission: ...


class ISecurityEncodable(typing.Protocol):
    @abc.abstractmethod
    def FromXml(self, e: SecurityElement) -> None: ...
    @abc.abstractmethod
    def ToXml(self) -> SecurityElement: ...


class IStackWalk(typing.Protocol):
    @abc.abstractmethod
    def Assert(self) -> None: ...
    @abc.abstractmethod
    def Demand(self) -> None: ...
    @abc.abstractmethod
    def Deny(self) -> None: ...
    @abc.abstractmethod
    def PermitOnly(self) -> None: ...


class PartialTrustVisibilityLevel(typing.SupportsInt):
    @typing.overload
    def __init__(self, value : int) -> None: ...
    @typing.overload
    def __init__(self, value : int, force_if_true: bool) -> None: ...
    def __int__(self) -> int: ...
    
    # Values:
    VisibleToAllHosts : PartialTrustVisibilityLevel # 0
    NotVisibleByDefault : PartialTrustVisibilityLevel # 1


class PermissionSet(ICollection, IStackWalk, ISecurityEncodable, IDeserializationCallback):
    @typing.overload
    def __init__(self, permSet: PermissionSet) -> None: ...
    @typing.overload
    def __init__(self, state: PermissionState) -> None: ...
    @property
    def Count(self) -> int: ...
    @property
    def IsReadOnly(self) -> bool: ...
    @property
    def IsSynchronized(self) -> bool: ...
    @property
    def SyncRoot(self) -> typing.Any: ...
    def AddPermission(self, perm: IPermission) -> IPermission: ...
    def Assert(self) -> None: ...
    def ContainsNonCodeAccessPermissions(self) -> bool: ...
    @staticmethod
    def ConvertPermissionSet(inFormat: str, inData: Array_1[int], outFormat: str) -> Array_1[int]: ...
    def Copy(self) -> PermissionSet: ...
    def CopyTo(self, array: Array, index: int) -> None: ...
    def Demand(self) -> None: ...
    def Deny(self) -> None: ...
    def Equals(self, o: typing.Any) -> bool: ...
    def FromXml(self, et: SecurityElement) -> None: ...
    def GetEnumerator(self) -> IEnumerator: ...
    def GetHashCode(self) -> int: ...
    def GetPermission(self, permClass: typing.Type[typing.Any]) -> IPermission: ...
    def Intersect(self, other: PermissionSet) -> PermissionSet: ...
    def IsEmpty(self) -> bool: ...
    def IsSubsetOf(self, target: PermissionSet) -> bool: ...
    def IsUnrestricted(self) -> bool: ...
    def PermitOnly(self) -> None: ...
    def RemovePermission(self, permClass: typing.Type[typing.Any]) -> IPermission: ...
    @staticmethod
    def RevertAssert() -> None: ...
    def SetPermission(self, perm: IPermission) -> IPermission: ...
    def ToString(self) -> str: ...
    def ToXml(self) -> SecurityElement: ...
    def Union(self, other: PermissionSet) -> PermissionSet: ...


class SecureString(IDisposable):
    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, value: clr.Reference[str], length: int) -> None: ...
    @property
    def Length(self) -> int: ...
    def AppendChar(self, c: str) -> None: ...
    def Clear(self) -> None: ...
    def Copy(self) -> SecureString: ...
    def Dispose(self) -> None: ...
    def InsertAt(self, index: int, c: str) -> None: ...
    def IsReadOnly(self) -> bool: ...
    def MakeReadOnly(self) -> None: ...
    def RemoveAt(self, index: int) -> None: ...
    def SetAt(self, index: int, c: str) -> None: ...


class SecurityCriticalAttribute(Attribute):
    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, scope: SecurityCriticalScope) -> None: ...
    @property
    def Scope(self) -> SecurityCriticalScope: ...
    @property
    def TypeId(self) -> typing.Any: ...


class SecurityCriticalScope(typing.SupportsInt):
    @typing.overload
    def __init__(self, value : int) -> None: ...
    @typing.overload
    def __init__(self, value : int, force_if_true: bool) -> None: ...
    def __int__(self) -> int: ...
    
    # Values:
    Explicit : SecurityCriticalScope # 0
    Everything : SecurityCriticalScope # 1


class SecurityElement:
    @typing.overload
    def __init__(self, tag: str) -> None: ...
    @typing.overload
    def __init__(self, tag: str, text: str) -> None: ...
    @property
    def Attributes(self) -> Hashtable: ...
    @Attributes.setter
    def Attributes(self, value: Hashtable) -> Hashtable: ...
    @property
    def Children(self) -> ArrayList: ...
    @Children.setter
    def Children(self, value: ArrayList) -> ArrayList: ...
    @property
    def Tag(self) -> str: ...
    @Tag.setter
    def Tag(self, value: str) -> str: ...
    @property
    def Text(self) -> str: ...
    @Text.setter
    def Text(self, value: str) -> str: ...
    def AddAttribute(self, name: str, value: str) -> None: ...
    def AddChild(self, child: SecurityElement) -> None: ...
    def Attribute(self, name: str) -> str: ...
    def Copy(self) -> SecurityElement: ...
    def Equal(self, other: SecurityElement) -> bool: ...
    @staticmethod
    def Escape(str: str) -> str: ...
    @staticmethod
    def FromString(xml: str) -> SecurityElement: ...
    @staticmethod
    def IsValidAttributeName(name: str) -> bool: ...
    @staticmethod
    def IsValidAttributeValue(value: str) -> bool: ...
    @staticmethod
    def IsValidTag(tag: str) -> bool: ...
    @staticmethod
    def IsValidText(text: str) -> bool: ...
    def SearchForChildByTag(self, tag: str) -> SecurityElement: ...
    def SearchForTextOfTag(self, tag: str) -> str: ...
    def ToString(self) -> str: ...


class SecurityException(SystemException):
    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, message: str) -> None: ...
    @typing.overload
    def __init__(self, message: str, inner: Exception) -> None: ...
    @typing.overload
    def __init__(self, message: str, type: typing.Type[typing.Any]) -> None: ...
    @typing.overload
    def __init__(self, message: str, type: typing.Type[typing.Any], state: str) -> None: ...
    @property
    def Data(self) -> IDictionary: ...
    @property
    def Demanded(self) -> typing.Any: ...
    @Demanded.setter
    def Demanded(self, value: typing.Any) -> typing.Any: ...
    @property
    def DenySetInstance(self) -> typing.Any: ...
    @DenySetInstance.setter
    def DenySetInstance(self, value: typing.Any) -> typing.Any: ...
    @property
    def FailedAssemblyInfo(self) -> AssemblyName: ...
    @FailedAssemblyInfo.setter
    def FailedAssemblyInfo(self, value: AssemblyName) -> AssemblyName: ...
    @property
    def GrantedSet(self) -> str: ...
    @GrantedSet.setter
    def GrantedSet(self, value: str) -> str: ...
    @property
    def HelpLink(self) -> str: ...
    @HelpLink.setter
    def HelpLink(self, value: str) -> str: ...
    @property
    def HResult(self) -> int: ...
    @HResult.setter
    def HResult(self, value: int) -> int: ...
    @property
    def InnerException(self) -> Exception: ...
    @property
    def Message(self) -> str: ...
    @property
    def Method(self) -> MethodInfo: ...
    @Method.setter
    def Method(self, value: MethodInfo) -> MethodInfo: ...
    @property
    def PermissionState(self) -> str: ...
    @PermissionState.setter
    def PermissionState(self, value: str) -> str: ...
    @property
    def PermissionType(self) -> typing.Type[typing.Any]: ...
    @PermissionType.setter
    def PermissionType(self, value: typing.Type[typing.Any]) -> typing.Type[typing.Any]: ...
    @property
    def PermitOnlySetInstance(self) -> typing.Any: ...
    @PermitOnlySetInstance.setter
    def PermitOnlySetInstance(self, value: typing.Any) -> typing.Any: ...
    @property
    def RefusedSet(self) -> str: ...
    @RefusedSet.setter
    def RefusedSet(self, value: str) -> str: ...
    @property
    def Source(self) -> str: ...
    @Source.setter
    def Source(self, value: str) -> str: ...
    @property
    def StackTrace(self) -> str: ...
    @property
    def TargetSite(self) -> MethodBase: ...
    @property
    def Url(self) -> str: ...
    @Url.setter
    def Url(self, value: str) -> str: ...
    def GetObjectData(self, info: SerializationInfo, context: StreamingContext) -> None: ...
    def ToString(self) -> str: ...


class SecurityRulesAttribute(Attribute):
    def __init__(self, ruleSet: SecurityRuleSet) -> None: ...
    @property
    def RuleSet(self) -> SecurityRuleSet: ...
    @property
    def SkipVerificationInFullTrust(self) -> bool: ...
    @SkipVerificationInFullTrust.setter
    def SkipVerificationInFullTrust(self, value: bool) -> bool: ...
    @property
    def TypeId(self) -> typing.Any: ...


class SecurityRuleSet(typing.SupportsInt):
    @typing.overload
    def __init__(self, value : int) -> None: ...
    @typing.overload
    def __init__(self, value : int, force_if_true: bool) -> None: ...
    def __int__(self) -> int: ...
    
    # Values:
    None_ : SecurityRuleSet # 0
    Level1 : SecurityRuleSet # 1
    Level2 : SecurityRuleSet # 2


class SecuritySafeCriticalAttribute(Attribute):
    def __init__(self) -> None: ...
    @property
    def TypeId(self) -> typing.Any: ...


class SecurityTransparentAttribute(Attribute):
    def __init__(self) -> None: ...
    @property
    def TypeId(self) -> typing.Any: ...


class SecurityTreatAsSafeAttribute(Attribute):
    def __init__(self) -> None: ...
    @property
    def TypeId(self) -> typing.Any: ...


class SuppressUnmanagedCodeSecurityAttribute(Attribute):
    def __init__(self) -> None: ...
    @property
    def TypeId(self) -> typing.Any: ...


class UnverifiableCodeAttribute(Attribute):
    def __init__(self) -> None: ...
    @property
    def TypeId(self) -> typing.Any: ...


class VerificationException(SystemException):
    @typing.overload
    def __init__(self) -> None: ...
    @typing.overload
    def __init__(self, message: str) -> None: ...
    @typing.overload
    def __init__(self, message: str, innerException: Exception) -> None: ...
    @property
    def Data(self) -> IDictionary: ...
    @property
    def HelpLink(self) -> str: ...
    @HelpLink.setter
    def HelpLink(self, value: str) -> str: ...
    @property
    def HResult(self) -> int: ...
    @HResult.setter
    def HResult(self, value: int) -> int: ...
    @property
    def InnerException(self) -> Exception: ...
    @property
    def Message(self) -> str: ...
    @property
    def Source(self) -> str: ...
    @Source.setter
    def Source(self, value: str) -> str: ...
    @property
    def StackTrace(self) -> str: ...
    @property
    def TargetSite(self) -> MethodBase: ...




from PySimultan2.taxonomy_maps import TaxonomyMap, Content
from PySimultan2.object_mapper import register
from PySimultan2 import PythonMapper


content_a = Content(text_or_key='a',
                    property_name='a',
                    type=None,
                    unit=None,
                    documentation='Property a',)

content_b = Content(text_or_key='b',
                    property_name='b',
                    type=None,
                    unit=None,
                    documentation='Property a',)

test_map = TaxonomyMap(taxonomy_name='Test',
                       taxonomy_key='Test',
                       taxonomy_entry_name='TestClass',
                       taxonomy_entry_key='TestClass',
                       content=[content_a, content_b],
                       )

@register(test_map.taxonomy_entry_key, taxonomy_map=test_map, re_register=True, module='test_mapper')
class TestClass(object):

    def __init__(self, *args, **kwargs):
        self.a = kwargs.get('a', 1)
        self.b = kwargs.get('b', 2)

    def add(self):
        return self.a + self.b

    def __repr__(self):
        return f'TestClass(a={self.a}, b={self.b})'



mapped_test_cls = PythonMapper.mappers['test_mapper'].get_mapped_class('TestClass')
print(mapped_test_cls)



import typing, clr, abc
from System.Reflection import MethodInfo, PropertyInfo, MemberInfo, FieldInfo, ConstructorInfo
from System.Collections.ObjectModel import ReadOnlyCollection_1
from System.Collections.Generic import IEnumerable_1
from System.Runtime.CompilerServices import CallSiteBinder, DebugInfoGenerator
from System import Array_1, Guid, Delegate

class BinaryExpression(Expression_0):
    @property
    def CanReduce(self) -> bool: ...
    @property
    def Conversion(self) -> LambdaExpression: ...
    @property
    def IsLifted(self) -> bool: ...
    @property
    def IsLiftedToNull(self) -> bool: ...
    @property
    def Left(self) -> Expression: ...
    @property
    def Method(self) -> MethodInfo: ...
    @property
    def NodeType(self) -> ExpressionType: ...
    @property
    def Right(self) -> Expression: ...
    @property
    def Type(self) -> typing.Type[typing.Any]: ...
    def Reduce(self) -> Expression: ...
    def Update(self, left: Expression, conversion: LambdaExpression, right: Expression) -> BinaryExpression: ...


class BlockExpression(Expression_0):
    @property
    def CanReduce(self) -> bool: ...
    @property
    def Expressions(self) -> ReadOnlyCollection_1[Expression]: ...
    @property
    def NodeType(self) -> ExpressionType: ...
    @property
    def Result(self) -> Expression: ...
    @property
    def Type(self) -> typing.Type[typing.Any]: ...
    @property
    def Variables(self) -> ReadOnlyCollection_1[ParameterExpression]: ...
    def Update(self, variables: IEnumerable_1[ParameterExpression], expressions: IEnumerable_1[Expression]) -> BlockExpression: ...


class CatchBlock:
    @property
    def Body(self) -> Expression: ...
    @property
    def Filter(self) -> Expression: ...
    @property
    def Test(self) -> typing.Type[typing.Any]: ...
    @property
    def Variable(self) -> ParameterExpression: ...
    def ToString(self) -> str: ...
    def Update(self, variable: ParameterExpression, filter: Expression, body: Expression) -> CatchBlock: ...


class ConditionalExpression(Expression_0):
    @property
    def CanReduce(self) -> bool: ...
    @property
    def IfFalse(self) -> Expression: ...
    @property
    def IfTrue(self) -> Expression: ...
    @property
    def NodeType(self) -> ExpressionType: ...
    @property
    def Test(self) -> Expression: ...
    @property
    def Type(self) -> typing.Type[typing.Any]: ...
    def Update(self, test: Expression, ifTrue: Expression, ifFalse: Expression) -> ConditionalExpression: ...


class ConstantExpression(Expression_0):
    @property
    def CanReduce(self) -> bool: ...
    @property
    def NodeType(self) -> ExpressionType: ...
    @property
    def Type(self) -> typing.Type[typing.Any]: ...
    @property
    def Value(self) -> typing.Any: ...


class DebugInfoExpression(Expression_0):
    @property
    def CanReduce(self) -> bool: ...
    @property
    def Document(self) -> SymbolDocumentInfo: ...
    @property
    def EndColumn(self) -> int: ...
    @property
    def EndLine(self) -> int: ...
    @property
    def IsClear(self) -> bool: ...
    @property
    def NodeType(self) -> ExpressionType: ...
    @property
    def StartColumn(self) -> int: ...
    @property
    def StartLine(self) -> int: ...
    @property
    def Type(self) -> typing.Type[typing.Any]: ...


class DefaultExpression(Expression_0):
    @property
    def CanReduce(self) -> bool: ...
    @property
    def NodeType(self) -> ExpressionType: ...
    @property
    def Type(self) -> typing.Type[typing.Any]: ...


class DynamicExpression(Expression_0, IDynamicExpression):
    @property
    def Arguments(self) -> ReadOnlyCollection_1[Expression]: ...
    @property
    def Binder(self) -> CallSiteBinder: ...
    @property
    def CanReduce(self) -> bool: ...
    @property
    def DelegateType(self) -> typing.Type[typing.Any]: ...
    @property
    def NodeType(self) -> ExpressionType: ...
    @property
    def Type(self) -> typing.Type[typing.Any]: ...
    def Reduce(self) -> Expression: ...
    def Update(self, arguments: IEnumerable_1[Expression]) -> DynamicExpression: ...
    # Skipped Dynamic due to it being static, abstract and generic.

    Dynamic : Dynamic_MethodGroup
    class Dynamic_MethodGroup:
        @typing.overload
        def __call__(self, binder: CallSiteBinder, returnType: typing.Type[typing.Any], arguments: Array_1[Expression]) -> DynamicExpression:...
        @typing.overload
        def __call__(self, binder: CallSiteBinder, returnType: typing.Type[typing.Any], arguments: IEnumerable_1[Expression]) -> DynamicExpression:...
        @typing.overload
        def __call__(self, binder: CallSiteBinder, returnType: typing.Type[typing.Any], arg0: Expression) -> DynamicExpression:...
        @typing.overload
        def __call__(self, binder: CallSiteBinder, returnType: typing.Type[typing.Any], arg0: Expression, arg1: Expression) -> DynamicExpression:...
        @typing.overload
        def __call__(self, binder: CallSiteBinder, returnType: typing.Type[typing.Any], arg0: Expression, arg1: Expression, arg2: Expression) -> DynamicExpression:...
        @typing.overload
        def __call__(self, binder: CallSiteBinder, returnType: typing.Type[typing.Any], arg0: Expression, arg1: Expression, arg2: Expression, arg3: Expression) -> DynamicExpression:...

    # Skipped MakeDynamic due to it being static, abstract and generic.

    MakeDynamic : MakeDynamic_MethodGroup
    class MakeDynamic_MethodGroup:
        @typing.overload
        def __call__(self, delegateType: typing.Type[typing.Any], binder: CallSiteBinder, arguments: Array_1[Expression]) -> DynamicExpression:...
        @typing.overload
        def __call__(self, delegateType: typing.Type[typing.Any], binder: CallSiteBinder, arguments: IEnumerable_1[Expression]) -> DynamicExpression:...
        @typing.overload
        def __call__(self, delegateType: typing.Type[typing.Any], binder: CallSiteBinder, arg0: Expression) -> DynamicExpression:...
        @typing.overload
        def __call__(self, delegateType: typing.Type[typing.Any], binder: CallSiteBinder, arg0: Expression, arg1: Expression) -> DynamicExpression:...
        @typing.overload
        def __call__(self, delegateType: typing.Type[typing.Any], binder: CallSiteBinder, arg0: Expression, arg1: Expression, arg2: Expression) -> DynamicExpression:...
        @typing.overload
        def __call__(self, delegateType: typing.Type[typing.Any], binder: CallSiteBinder, arg0: Expression, arg1: Expression, arg2: Expression, arg3: Expression) -> DynamicExpression:...



class ElementInit(IArgumentProvider):
    @property
    def AddMethod(self) -> MethodInfo: ...
    @property
    def ArgumentCount(self) -> int: ...
    @property
    def Arguments(self) -> ReadOnlyCollection_1[Expression]: ...
    def GetArgument(self, index: int) -> Expression: ...
    def ToString(self) -> str: ...
    def Update(self, arguments: IEnumerable_1[Expression]) -> ElementInit: ...


class Expression_GenericClasses(abc.ABCMeta):
    Generic_Expression_GenericClasses_Expression_1_TDelegate = typing.TypeVar('Generic_Expression_GenericClasses_Expression_1_TDelegate')
    def __getitem__(self, types : typing.Type[Generic_Expression_GenericClasses_Expression_1_TDelegate]) -> typing.Type[Expression_1[Generic_Expression_GenericClasses_Expression_1_TDelegate]]: ...

class Expression(Expression_0, metaclass =Expression_GenericClasses): ...

class Expression_0(abc.ABC):
    @property
    def CanReduce(self) -> bool: ...
    @property
    def NodeType(self) -> ExpressionType: ...
    @property
    def Type(self) -> typing.Type[typing.Any]: ...
    @staticmethod
    def ArrayLength(array: Expression) -> UnaryExpression: ...
    @staticmethod
    def Assign(left: Expression, right: Expression) -> BinaryExpression: ...
    @staticmethod
    def ClearDebugInfo(document: SymbolDocumentInfo) -> DebugInfoExpression: ...
    @staticmethod
    def DebugInfo(document: SymbolDocumentInfo, startLine: int, startColumn: int, endLine: int, endColumn: int) -> DebugInfoExpression: ...
    @staticmethod
    def Default(type: typing.Type[typing.Any]) -> DefaultExpression: ...
    @staticmethod
    def Empty() -> DefaultExpression: ...
    @staticmethod
    def GetActionType(typeArgs: Array_1[typing.Type[typing.Any]]) -> typing.Type[typing.Any]: ...
    @staticmethod
    def GetDelegateType(typeArgs: Array_1[typing.Type[typing.Any]]) -> typing.Type[typing.Any]: ...
    @staticmethod
    def GetFuncType(typeArgs: Array_1[typing.Type[typing.Any]]) -> typing.Type[typing.Any]: ...
    @staticmethod
    def IfThen(test: Expression, ifTrue: Expression) -> ConditionalExpression: ...
    @staticmethod
    def IfThenElse(test: Expression, ifTrue: Expression, ifFalse: Expression) -> ConditionalExpression: ...
    @staticmethod
    def MakeCatchBlock(type: typing.Type[typing.Any], variable: ParameterExpression, body: Expression, filter: Expression) -> CatchBlock: ...
    @staticmethod
    def MakeGoto(kind: GotoExpressionKind, target: LabelTarget, value: Expression, type: typing.Type[typing.Any]) -> GotoExpression: ...
    @staticmethod
    def MakeIndex(instance: Expression, indexer: PropertyInfo, arguments: IEnumerable_1[Expression]) -> IndexExpression: ...
    @staticmethod
    def MakeMemberAccess(expression: Expression, member: MemberInfo) -> MemberExpression: ...
    @staticmethod
    def MakeTry(type: typing.Type[typing.Any], body: Expression, finally: Expression, fault: Expression, handlers: IEnumerable_1[CatchBlock]) -> TryExpression: ...
    @staticmethod
    def PropertyOrField(expression: Expression, propertyOrFieldName: str) -> MemberExpression: ...
    @staticmethod
    def Quote(expression: Expression) -> UnaryExpression: ...
    def Reduce(self) -> Expression: ...
    def ReduceAndCheck(self) -> Expression: ...
    def ReduceExtensions(self) -> Expression: ...
    @staticmethod
    def ReferenceEqual(left: Expression, right: Expression) -> BinaryExpression: ...
    @staticmethod
    def ReferenceNotEqual(left: Expression, right: Expression) -> BinaryExpression: ...
    def ToString(self) -> str: ...
    @staticmethod
    def TryCatch(body: Expression, handlers: Array_1[CatchBlock]) -> TryExpression: ...
    @staticmethod
    def TryCatchFinally(body: Expression, finally: Expression, handlers: Array_1[CatchBlock]) -> TryExpression: ...
    @staticmethod
    def TryFault(body: Expression, fault: Expression) -> TryExpression: ...
    @staticmethod
    def TryFinally(body: Expression, finally: Expression) -> TryExpression: ...
    @staticmethod
    def TryGetActionType(typeArgs: Array_1[typing.Type[typing.Any]], actionType: clr.Reference[typing.Type[typing.Any]]) -> bool: ...
    @staticmethod
    def TryGetFuncType(typeArgs: Array_1[typing.Type[typing.Any]], funcType: clr.Reference[typing.Type[typing.Any]]) -> bool: ...
    @staticmethod
    def TypeAs(expression: Expression, type: typing.Type[typing.Any]) -> UnaryExpression: ...
    @staticmethod
    def TypeEqual(expression: Expression, type: typing.Type[typing.Any]) -> TypeBinaryExpression: ...
    @staticmethod
    def TypeIs(expression: Expression, type: typing.Type[typing.Any]) -> TypeBinaryExpression: ...
    @staticmethod
    def Unbox(expression: Expression, type: typing.Type[typing.Any]) -> UnaryExpression: ...
    # Skipped Add due to it being static, abstract and generic.

    Add : Add_MethodGroup
    class Add_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo) -> BinaryExpression:...

    # Skipped AddAssign due to it being static, abstract and generic.

    AddAssign : AddAssign_MethodGroup
    class AddAssign_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo, conversion: LambdaExpression) -> BinaryExpression:...

    # Skipped AddAssignChecked due to it being static, abstract and generic.

    AddAssignChecked : AddAssignChecked_MethodGroup
    class AddAssignChecked_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo, conversion: LambdaExpression) -> BinaryExpression:...

    # Skipped AddChecked due to it being static, abstract and generic.

    AddChecked : AddChecked_MethodGroup
    class AddChecked_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo) -> BinaryExpression:...

    # Skipped And due to it being static, abstract and generic.

    And : And_MethodGroup
    class And_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo) -> BinaryExpression:...

    # Skipped AndAlso due to it being static, abstract and generic.

    AndAlso : AndAlso_MethodGroup
    class AndAlso_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo) -> BinaryExpression:...

    # Skipped AndAssign due to it being static, abstract and generic.

    AndAssign : AndAssign_MethodGroup
    class AndAssign_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo, conversion: LambdaExpression) -> BinaryExpression:...

    # Skipped ArrayAccess due to it being static, abstract and generic.

    ArrayAccess : ArrayAccess_MethodGroup
    class ArrayAccess_MethodGroup:
        @typing.overload
        def __call__(self, array: Expression, indexes: Array_1[Expression]) -> IndexExpression:...
        @typing.overload
        def __call__(self, array: Expression, indexes: IEnumerable_1[Expression]) -> IndexExpression:...

    # Skipped ArrayIndex due to it being static, abstract and generic.

    ArrayIndex : ArrayIndex_MethodGroup
    class ArrayIndex_MethodGroup:
        @typing.overload
        def __call__(self, array: Expression, indexes: Array_1[Expression]) -> MethodCallExpression:...
        @typing.overload
        def __call__(self, array: Expression, indexes: IEnumerable_1[Expression]) -> MethodCallExpression:...
        @typing.overload
        def __call__(self, array: Expression, index: Expression) -> BinaryExpression:...

    # Skipped Bind due to it being static, abstract and generic.

    Bind : Bind_MethodGroup
    class Bind_MethodGroup:
        @typing.overload
        def __call__(self, propertyAccessor: MethodInfo, expression: Expression) -> MemberAssignment:...
        @typing.overload
        def __call__(self, member: MemberInfo, expression: Expression) -> MemberAssignment:...

    # Skipped Block due to it being static, abstract and generic.

    Block : Block_MethodGroup
    class Block_MethodGroup:
        @typing.overload
        def __call__(self, expressions: Array_1[Expression]) -> BlockExpression:...
        @typing.overload
        def __call__(self, expressions: IEnumerable_1[Expression]) -> BlockExpression:...
        @typing.overload
        def __call__(self, type: typing.Type[typing.Any], expressions: Array_1[Expression]) -> BlockExpression:...
        @typing.overload
        def __call__(self, type: typing.Type[typing.Any], expressions: IEnumerable_1[Expression]) -> BlockExpression:...
        @typing.overload
        def __call__(self, variables: IEnumerable_1[ParameterExpression], expressions: Array_1[Expression]) -> BlockExpression:...
        @typing.overload
        def __call__(self, variables: IEnumerable_1[ParameterExpression], expressions: IEnumerable_1[Expression]) -> BlockExpression:...
        @typing.overload
        def __call__(self, arg0: Expression, arg1: Expression) -> BlockExpression:...
        @typing.overload
        def __call__(self, type: typing.Type[typing.Any], variables: IEnumerable_1[ParameterExpression], expressions: Array_1[Expression]) -> BlockExpression:...
        @typing.overload
        def __call__(self, type: typing.Type[typing.Any], variables: IEnumerable_1[ParameterExpression], expressions: IEnumerable_1[Expression]) -> BlockExpression:...
        @typing.overload
        def __call__(self, arg0: Expression, arg1: Expression, arg2: Expression) -> BlockExpression:...
        @typing.overload
        def __call__(self, arg0: Expression, arg1: Expression, arg2: Expression, arg3: Expression) -> BlockExpression:...
        @typing.overload
        def __call__(self, arg0: Expression, arg1: Expression, arg2: Expression, arg3: Expression, arg4: Expression) -> BlockExpression:...

    # Skipped Break due to it being static, abstract and generic.

    Break : Break_MethodGroup
    class Break_MethodGroup:
        @typing.overload
        def __call__(self, target: LabelTarget) -> GotoExpression:...
        @typing.overload
        def __call__(self, target: LabelTarget, type: typing.Type[typing.Any]) -> GotoExpression:...
        @typing.overload
        def __call__(self, target: LabelTarget, value: Expression) -> GotoExpression:...
        @typing.overload
        def __call__(self, target: LabelTarget, value: Expression, type: typing.Type[typing.Any]) -> GotoExpression:...

    # Skipped Call due to it being static, abstract and generic.

    Call : Call_MethodGroup
    class Call_MethodGroup:
        @typing.overload
        def __call__(self, method: MethodInfo, arguments: Array_1[Expression]) -> MethodCallExpression:...
        @typing.overload
        def __call__(self, method: MethodInfo, arguments: IEnumerable_1[Expression]) -> MethodCallExpression:...
        @typing.overload
        def __call__(self, method: MethodInfo, arg0: Expression) -> MethodCallExpression:...
        @typing.overload
        def __call__(self, instance: Expression, method: MethodInfo) -> MethodCallExpression:...
        @typing.overload
        def __call__(self, method: MethodInfo, arg0: Expression, arg1: Expression) -> MethodCallExpression:...
        @typing.overload
        def __call__(self, instance: Expression, method: MethodInfo, arguments: Array_1[Expression]) -> MethodCallExpression:...
        @typing.overload
        def __call__(self, instance: Expression, method: MethodInfo, arguments: IEnumerable_1[Expression]) -> MethodCallExpression:...
        @typing.overload
        def __call__(self, method: MethodInfo, arg0: Expression, arg1: Expression, arg2: Expression) -> MethodCallExpression:...
        @typing.overload
        def __call__(self, type: typing.Type[typing.Any], methodName: str, typeArguments: Array_1[typing.Type[typing.Any]], arguments: Array_1[Expression]) -> MethodCallExpression:...
        @typing.overload
        def __call__(self, instance: Expression, method: MethodInfo, arg0: Expression, arg1: Expression) -> MethodCallExpression:...
        @typing.overload
        def __call__(self, instance: Expression, methodName: str, typeArguments: Array_1[typing.Type[typing.Any]], arguments: Array_1[Expression]) -> MethodCallExpression:...
        @typing.overload
        def __call__(self, method: MethodInfo, arg0: Expression, arg1: Expression, arg2: Expression, arg3: Expression) -> MethodCallExpression:...
        @typing.overload
        def __call__(self, instance: Expression, method: MethodInfo, arg0: Expression, arg1: Expression, arg2: Expression) -> MethodCallExpression:...
        @typing.overload
        def __call__(self, method: MethodInfo, arg0: Expression, arg1: Expression, arg2: Expression, arg3: Expression, arg4: Expression) -> MethodCallExpression:...

    # Skipped Catch due to it being static, abstract and generic.

    Catch : Catch_MethodGroup
    class Catch_MethodGroup:
        @typing.overload
        def __call__(self, type: typing.Type[typing.Any], body: Expression) -> CatchBlock:...
        @typing.overload
        def __call__(self, variable: ParameterExpression, body: Expression) -> CatchBlock:...
        @typing.overload
        def __call__(self, type: typing.Type[typing.Any], body: Expression, filter: Expression) -> CatchBlock:...
        @typing.overload
        def __call__(self, variable: ParameterExpression, body: Expression, filter: Expression) -> CatchBlock:...

    # Skipped Coalesce due to it being static, abstract and generic.

    Coalesce : Coalesce_MethodGroup
    class Coalesce_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, conversion: LambdaExpression) -> BinaryExpression:...

    # Skipped Condition due to it being static, abstract and generic.

    Condition : Condition_MethodGroup
    class Condition_MethodGroup:
        @typing.overload
        def __call__(self, test: Expression, ifTrue: Expression, ifFalse: Expression) -> ConditionalExpression:...
        @typing.overload
        def __call__(self, test: Expression, ifTrue: Expression, ifFalse: Expression, type: typing.Type[typing.Any]) -> ConditionalExpression:...

    # Skipped Constant due to it being static, abstract and generic.

    Constant : Constant_MethodGroup
    class Constant_MethodGroup:
        @typing.overload
        def __call__(self, value: typing.Any) -> ConstantExpression:...
        @typing.overload
        def __call__(self, value: typing.Any, type: typing.Type[typing.Any]) -> ConstantExpression:...

    # Skipped Continue due to it being static, abstract and generic.

    Continue : Continue_MethodGroup
    class Continue_MethodGroup:
        @typing.overload
        def __call__(self, target: LabelTarget) -> GotoExpression:...
        @typing.overload
        def __call__(self, target: LabelTarget, type: typing.Type[typing.Any]) -> GotoExpression:...

    # Skipped Convert due to it being static, abstract and generic.

    Convert : Convert_MethodGroup
    class Convert_MethodGroup:
        @typing.overload
        def __call__(self, expression: Expression, type: typing.Type[typing.Any]) -> UnaryExpression:...
        @typing.overload
        def __call__(self, expression: Expression, type: typing.Type[typing.Any], method: MethodInfo) -> UnaryExpression:...

    # Skipped ConvertChecked due to it being static, abstract and generic.

    ConvertChecked : ConvertChecked_MethodGroup
    class ConvertChecked_MethodGroup:
        @typing.overload
        def __call__(self, expression: Expression, type: typing.Type[typing.Any]) -> UnaryExpression:...
        @typing.overload
        def __call__(self, expression: Expression, type: typing.Type[typing.Any], method: MethodInfo) -> UnaryExpression:...

    # Skipped Decrement due to it being static, abstract and generic.

    Decrement : Decrement_MethodGroup
    class Decrement_MethodGroup:
        @typing.overload
        def __call__(self, expression: Expression) -> UnaryExpression:...
        @typing.overload
        def __call__(self, expression: Expression, method: MethodInfo) -> UnaryExpression:...

    # Skipped Divide due to it being static, abstract and generic.

    Divide : Divide_MethodGroup
    class Divide_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo) -> BinaryExpression:...

    # Skipped DivideAssign due to it being static, abstract and generic.

    DivideAssign : DivideAssign_MethodGroup
    class DivideAssign_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo, conversion: LambdaExpression) -> BinaryExpression:...

    # Skipped Dynamic due to it being static, abstract and generic.

    Dynamic : Dynamic_MethodGroup
    class Dynamic_MethodGroup:
        @typing.overload
        def __call__(self, binder: CallSiteBinder, returnType: typing.Type[typing.Any], arguments: Array_1[Expression]) -> DynamicExpression:...
        @typing.overload
        def __call__(self, binder: CallSiteBinder, returnType: typing.Type[typing.Any], arguments: IEnumerable_1[Expression]) -> DynamicExpression:...
        @typing.overload
        def __call__(self, binder: CallSiteBinder, returnType: typing.Type[typing.Any], arg0: Expression) -> DynamicExpression:...
        @typing.overload
        def __call__(self, binder: CallSiteBinder, returnType: typing.Type[typing.Any], arg0: Expression, arg1: Expression) -> DynamicExpression:...
        @typing.overload
        def __call__(self, binder: CallSiteBinder, returnType: typing.Type[typing.Any], arg0: Expression, arg1: Expression, arg2: Expression) -> DynamicExpression:...
        @typing.overload
        def __call__(self, binder: CallSiteBinder, returnType: typing.Type[typing.Any], arg0: Expression, arg1: Expression, arg2: Expression, arg3: Expression) -> DynamicExpression:...

    # Skipped ElementInit due to it being static, abstract and generic.

    ElementInit : ElementInit_MethodGroup
    class ElementInit_MethodGroup:
        @typing.overload
        def __call__(self, addMethod: MethodInfo, arguments: Array_1[Expression]) -> ElementInit:...
        @typing.overload
        def __call__(self, addMethod: MethodInfo, arguments: IEnumerable_1[Expression]) -> ElementInit:...

    # Skipped Equal due to it being static, abstract and generic.

    Equal : Equal_MethodGroup
    class Equal_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, liftToNull: bool, method: MethodInfo) -> BinaryExpression:...

    # Skipped ExclusiveOr due to it being static, abstract and generic.

    ExclusiveOr : ExclusiveOr_MethodGroup
    class ExclusiveOr_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo) -> BinaryExpression:...

    # Skipped ExclusiveOrAssign due to it being static, abstract and generic.

    ExclusiveOrAssign : ExclusiveOrAssign_MethodGroup
    class ExclusiveOrAssign_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo, conversion: LambdaExpression) -> BinaryExpression:...

    # Skipped Field due to it being static, abstract and generic.

    Field : Field_MethodGroup
    class Field_MethodGroup:
        @typing.overload
        def __call__(self, expression: Expression, field: FieldInfo) -> MemberExpression:...
        @typing.overload
        def __call__(self, expression: Expression, fieldName: str) -> MemberExpression:...
        @typing.overload
        def __call__(self, expression: Expression, type: typing.Type[typing.Any], fieldName: str) -> MemberExpression:...

    # Skipped Goto due to it being static, abstract and generic.

    Goto : Goto_MethodGroup
    class Goto_MethodGroup:
        @typing.overload
        def __call__(self, target: LabelTarget) -> GotoExpression:...
        @typing.overload
        def __call__(self, target: LabelTarget, type: typing.Type[typing.Any]) -> GotoExpression:...
        @typing.overload
        def __call__(self, target: LabelTarget, value: Expression) -> GotoExpression:...
        @typing.overload
        def __call__(self, target: LabelTarget, value: Expression, type: typing.Type[typing.Any]) -> GotoExpression:...

    # Skipped GreaterThan due to it being static, abstract and generic.

    GreaterThan : GreaterThan_MethodGroup
    class GreaterThan_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, liftToNull: bool, method: MethodInfo) -> BinaryExpression:...

    # Skipped GreaterThanOrEqual due to it being static, abstract and generic.

    GreaterThanOrEqual : GreaterThanOrEqual_MethodGroup
    class GreaterThanOrEqual_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, liftToNull: bool, method: MethodInfo) -> BinaryExpression:...

    # Skipped Increment due to it being static, abstract and generic.

    Increment : Increment_MethodGroup
    class Increment_MethodGroup:
        @typing.overload
        def __call__(self, expression: Expression) -> UnaryExpression:...
        @typing.overload
        def __call__(self, expression: Expression, method: MethodInfo) -> UnaryExpression:...

    # Skipped Invoke due to it being static, abstract and generic.

    Invoke : Invoke_MethodGroup
    class Invoke_MethodGroup:
        @typing.overload
        def __call__(self, expression: Expression, arguments: Array_1[Expression]) -> InvocationExpression:...
        @typing.overload
        def __call__(self, expression: Expression, arguments: IEnumerable_1[Expression]) -> InvocationExpression:...

    # Skipped IsFalse due to it being static, abstract and generic.

    IsFalse : IsFalse_MethodGroup
    class IsFalse_MethodGroup:
        @typing.overload
        def __call__(self, expression: Expression) -> UnaryExpression:...
        @typing.overload
        def __call__(self, expression: Expression, method: MethodInfo) -> UnaryExpression:...

    # Skipped IsTrue due to it being static, abstract and generic.

    IsTrue : IsTrue_MethodGroup
    class IsTrue_MethodGroup:
        @typing.overload
        def __call__(self, expression: Expression) -> UnaryExpression:...
        @typing.overload
        def __call__(self, expression: Expression, method: MethodInfo) -> UnaryExpression:...

    # Skipped Label due to it being static, abstract and generic.

    Label : Label_MethodGroup
    class Label_MethodGroup:
        @typing.overload
        def __call__(self) -> LabelTarget:...
        @typing.overload
        def __call__(self, name: str) -> LabelTarget:...
        @typing.overload
        def __call__(self, type: typing.Type[typing.Any]) -> LabelTarget:...
        @typing.overload
        def __call__(self, target: LabelTarget) -> LabelExpression:...
        @typing.overload
        def __call__(self, type: typing.Type[typing.Any], name: str) -> LabelTarget:...
        @typing.overload
        def __call__(self, target: LabelTarget, defaultValue: Expression) -> LabelExpression:...

    # Skipped Lambda due to it being static, abstract and generic.

    Lambda : Lambda_MethodGroup
    class Lambda_MethodGroup:
        def __getitem__(self, t:typing.Type[Lambda_1_T1]) -> Lambda_1[Lambda_1_T1]: ...

        Lambda_1_T1 = typing.TypeVar('Lambda_1_T1')
        class Lambda_1(typing.Generic[Lambda_1_T1]):
            Lambda_1_TDelegate = Expression_0.Lambda_MethodGroup.Lambda_1_T1
            @typing.overload
            def __call__(self, body: Expression, parameters: Array_1[ParameterExpression]) -> Expression_1[Lambda_1_TDelegate]:...
            @typing.overload
            def __call__(self, body: Expression, parameters: IEnumerable_1[ParameterExpression]) -> Expression_1[Lambda_1_TDelegate]:...
            @typing.overload
            def __call__(self, body: Expression, tailCall: bool, parameters: Array_1[ParameterExpression]) -> Expression_1[Lambda_1_TDelegate]:...
            @typing.overload
            def __call__(self, body: Expression, name: str, parameters: IEnumerable_1[ParameterExpression]) -> Expression_1[Lambda_1_TDelegate]:...
            @typing.overload
            def __call__(self, body: Expression, tailCall: bool, parameters: IEnumerable_1[ParameterExpression]) -> Expression_1[Lambda_1_TDelegate]:...
            @typing.overload
            def __call__(self, body: Expression, name: str, tailCall: bool, parameters: IEnumerable_1[ParameterExpression]) -> Expression_1[Lambda_1_TDelegate]:...

        @typing.overload
        def __call__(self, body: Expression, parameters: Array_1[ParameterExpression]) -> LambdaExpression:...
        @typing.overload
        def __call__(self, body: Expression, parameters: IEnumerable_1[ParameterExpression]) -> LambdaExpression:...
        @typing.overload
        def __call__(self, delegateType: typing.Type[typing.Any], body: Expression, parameters: Array_1[ParameterExpression]) -> LambdaExpression:...
        @typing.overload
        def __call__(self, delegateType: typing.Type[typing.Any], body: Expression, parameters: IEnumerable_1[ParameterExpression]) -> LambdaExpression:...
        @typing.overload
        def __call__(self, body: Expression, tailCall: bool, parameters: Array_1[ParameterExpression]) -> LambdaExpression:...
        @typing.overload
        def __call__(self, body: Expression, name: str, parameters: IEnumerable_1[ParameterExpression]) -> LambdaExpression:...
        @typing.overload
        def __call__(self, body: Expression, tailCall: bool, parameters: IEnumerable_1[ParameterExpression]) -> LambdaExpression:...
        @typing.overload
        def __call__(self, delegateType: typing.Type[typing.Any], body: Expression, tailCall: bool, parameters: Array_1[ParameterExpression]) -> LambdaExpression:...
        @typing.overload
        def __call__(self, delegateType: typing.Type[typing.Any], body: Expression, name: str, parameters: IEnumerable_1[ParameterExpression]) -> LambdaExpression:...
        @typing.overload
        def __call__(self, delegateType: typing.Type[typing.Any], body: Expression, tailCall: bool, parameters: IEnumerable_1[ParameterExpression]) -> LambdaExpression:...
        @typing.overload
        def __call__(self, body: Expression, name: str, tailCall: bool, parameters: IEnumerable_1[ParameterExpression]) -> LambdaExpression:...
        @typing.overload
        def __call__(self, delegateType: typing.Type[typing.Any], body: Expression, name: str, tailCall: bool, parameters: IEnumerable_1[ParameterExpression]) -> LambdaExpression:...

    # Skipped LeftShift due to it being static, abstract and generic.

    LeftShift : LeftShift_MethodGroup
    class LeftShift_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo) -> BinaryExpression:...

    # Skipped LeftShiftAssign due to it being static, abstract and generic.

    LeftShiftAssign : LeftShiftAssign_MethodGroup
    class LeftShiftAssign_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo, conversion: LambdaExpression) -> BinaryExpression:...

    # Skipped LessThan due to it being static, abstract and generic.

    LessThan : LessThan_MethodGroup
    class LessThan_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, liftToNull: bool, method: MethodInfo) -> BinaryExpression:...

    # Skipped LessThanOrEqual due to it being static, abstract and generic.

    LessThanOrEqual : LessThanOrEqual_MethodGroup
    class LessThanOrEqual_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, liftToNull: bool, method: MethodInfo) -> BinaryExpression:...

    # Skipped ListBind due to it being static, abstract and generic.

    ListBind : ListBind_MethodGroup
    class ListBind_MethodGroup:
        @typing.overload
        def __call__(self, propertyAccessor: MethodInfo, initializers: Array_1[ElementInit]) -> MemberListBinding:...
        @typing.overload
        def __call__(self, propertyAccessor: MethodInfo, initializers: IEnumerable_1[ElementInit]) -> MemberListBinding:...
        @typing.overload
        def __call__(self, member: MemberInfo, initializers: Array_1[ElementInit]) -> MemberListBinding:...
        @typing.overload
        def __call__(self, member: MemberInfo, initializers: IEnumerable_1[ElementInit]) -> MemberListBinding:...

    # Skipped ListInit due to it being static, abstract and generic.

    ListInit : ListInit_MethodGroup
    class ListInit_MethodGroup:
        @typing.overload
        def __call__(self, newExpression: NewExpression, initializers: Array_1[Expression]) -> ListInitExpression:...
        @typing.overload
        def __call__(self, newExpression: NewExpression, initializers: Array_1[ElementInit]) -> ListInitExpression:...
        @typing.overload
        def __call__(self, newExpression: NewExpression, initializers: IEnumerable_1[Expression]) -> ListInitExpression:...
        @typing.overload
        def __call__(self, newExpression: NewExpression, initializers: IEnumerable_1[ElementInit]) -> ListInitExpression:...
        @typing.overload
        def __call__(self, newExpression: NewExpression, addMethod: MethodInfo, initializers: Array_1[Expression]) -> ListInitExpression:...
        @typing.overload
        def __call__(self, newExpression: NewExpression, addMethod: MethodInfo, initializers: IEnumerable_1[Expression]) -> ListInitExpression:...

    # Skipped Loop due to it being static, abstract and generic.

    Loop : Loop_MethodGroup
    class Loop_MethodGroup:
        @typing.overload
        def __call__(self, body: Expression) -> LoopExpression:...
        @typing.overload
        def __call__(self, body: Expression, break: LabelTarget) -> LoopExpression:...
        @typing.overload
        def __call__(self, body: Expression, break: LabelTarget, continue: LabelTarget) -> LoopExpression:...

    # Skipped MakeBinary due to it being static, abstract and generic.

    MakeBinary : MakeBinary_MethodGroup
    class MakeBinary_MethodGroup:
        @typing.overload
        def __call__(self, binaryType: ExpressionType, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, binaryType: ExpressionType, left: Expression, right: Expression, liftToNull: bool, method: MethodInfo) -> BinaryExpression:...
        @typing.overload
        def __call__(self, binaryType: ExpressionType, left: Expression, right: Expression, liftToNull: bool, method: MethodInfo, conversion: LambdaExpression) -> BinaryExpression:...

    # Skipped MakeDynamic due to it being static, abstract and generic.

    MakeDynamic : MakeDynamic_MethodGroup
    class MakeDynamic_MethodGroup:
        @typing.overload
        def __call__(self, delegateType: typing.Type[typing.Any], binder: CallSiteBinder, arguments: Array_1[Expression]) -> DynamicExpression:...
        @typing.overload
        def __call__(self, delegateType: typing.Type[typing.Any], binder: CallSiteBinder, arguments: IEnumerable_1[Expression]) -> DynamicExpression:...
        @typing.overload
        def __call__(self, delegateType: typing.Type[typing.Any], binder: CallSiteBinder, arg0: Expression) -> DynamicExpression:...
        @typing.overload
        def __call__(self, delegateType: typing.Type[typing.Any], binder: CallSiteBinder, arg0: Expression, arg1: Expression) -> DynamicExpression:...
        @typing.overload
        def __call__(self, delegateType: typing.Type[typing.Any], binder: CallSiteBinder, arg0: Expression, arg1: Expression, arg2: Expression) -> DynamicExpression:...
        @typing.overload
        def __call__(self, delegateType: typing.Type[typing.Any], binder: CallSiteBinder, arg0: Expression, arg1: Expression, arg2: Expression, arg3: Expression) -> DynamicExpression:...

    # Skipped MakeUnary due to it being static, abstract and generic.

    MakeUnary : MakeUnary_MethodGroup
    class MakeUnary_MethodGroup:
        @typing.overload
        def __call__(self, unaryType: ExpressionType, operand: Expression, type: typing.Type[typing.Any]) -> UnaryExpression:...
        @typing.overload
        def __call__(self, unaryType: ExpressionType, operand: Expression, type: typing.Type[typing.Any], method: MethodInfo) -> UnaryExpression:...

    # Skipped MemberBind due to it being static, abstract and generic.

    MemberBind : MemberBind_MethodGroup
    class MemberBind_MethodGroup:
        @typing.overload
        def __call__(self, propertyAccessor: MethodInfo, bindings: Array_1[MemberBinding]) -> MemberMemberBinding:...
        @typing.overload
        def __call__(self, propertyAccessor: MethodInfo, bindings: IEnumerable_1[MemberBinding]) -> MemberMemberBinding:...
        @typing.overload
        def __call__(self, member: MemberInfo, bindings: Array_1[MemberBinding]) -> MemberMemberBinding:...
        @typing.overload
        def __call__(self, member: MemberInfo, bindings: IEnumerable_1[MemberBinding]) -> MemberMemberBinding:...

    # Skipped MemberInit due to it being static, abstract and generic.

    MemberInit : MemberInit_MethodGroup
    class MemberInit_MethodGroup:
        @typing.overload
        def __call__(self, newExpression: NewExpression, bindings: Array_1[MemberBinding]) -> MemberInitExpression:...
        @typing.overload
        def __call__(self, newExpression: NewExpression, bindings: IEnumerable_1[MemberBinding]) -> MemberInitExpression:...

    # Skipped Modulo due to it being static, abstract and generic.

    Modulo : Modulo_MethodGroup
    class Modulo_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo) -> BinaryExpression:...

    # Skipped ModuloAssign due to it being static, abstract and generic.

    ModuloAssign : ModuloAssign_MethodGroup
    class ModuloAssign_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo, conversion: LambdaExpression) -> BinaryExpression:...

    # Skipped Multiply due to it being static, abstract and generic.

    Multiply : Multiply_MethodGroup
    class Multiply_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo) -> BinaryExpression:...

    # Skipped MultiplyAssign due to it being static, abstract and generic.

    MultiplyAssign : MultiplyAssign_MethodGroup
    class MultiplyAssign_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo, conversion: LambdaExpression) -> BinaryExpression:...

    # Skipped MultiplyAssignChecked due to it being static, abstract and generic.

    MultiplyAssignChecked : MultiplyAssignChecked_MethodGroup
    class MultiplyAssignChecked_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo, conversion: LambdaExpression) -> BinaryExpression:...

    # Skipped MultiplyChecked due to it being static, abstract and generic.

    MultiplyChecked : MultiplyChecked_MethodGroup
    class MultiplyChecked_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo) -> BinaryExpression:...

    # Skipped Negate due to it being static, abstract and generic.

    Negate : Negate_MethodGroup
    class Negate_MethodGroup:
        @typing.overload
        def __call__(self, expression: Expression) -> UnaryExpression:...
        @typing.overload
        def __call__(self, expression: Expression, method: MethodInfo) -> UnaryExpression:...

    # Skipped NegateChecked due to it being static, abstract and generic.

    NegateChecked : NegateChecked_MethodGroup
    class NegateChecked_MethodGroup:
        @typing.overload
        def __call__(self, expression: Expression) -> UnaryExpression:...
        @typing.overload
        def __call__(self, expression: Expression, method: MethodInfo) -> UnaryExpression:...

    # Skipped New due to it being static, abstract and generic.

    New : New_MethodGroup
    class New_MethodGroup:
        @typing.overload
        def __call__(self, constructor: ConstructorInfo) -> NewExpression:...
        @typing.overload
        def __call__(self, type: typing.Type[typing.Any]) -> NewExpression:...
        @typing.overload
        def __call__(self, constructor: ConstructorInfo, arguments: Array_1[Expression]) -> NewExpression:...
        @typing.overload
        def __call__(self, constructor: ConstructorInfo, arguments: IEnumerable_1[Expression]) -> NewExpression:...
        @typing.overload
        def __call__(self, constructor: ConstructorInfo, arguments: IEnumerable_1[Expression], members: Array_1[MemberInfo]) -> NewExpression:...
        @typing.overload
        def __call__(self, constructor: ConstructorInfo, arguments: IEnumerable_1[Expression], members: IEnumerable_1[MemberInfo]) -> NewExpression:...

    # Skipped NewArrayBounds due to it being static, abstract and generic.

    NewArrayBounds : NewArrayBounds_MethodGroup
    class NewArrayBounds_MethodGroup:
        @typing.overload
        def __call__(self, type: typing.Type[typing.Any], bounds: Array_1[Expression]) -> NewArrayExpression:...
        @typing.overload
        def __call__(self, type: typing.Type[typing.Any], bounds: IEnumerable_1[Expression]) -> NewArrayExpression:...

    # Skipped NewArrayInit due to it being static, abstract and generic.

    NewArrayInit : NewArrayInit_MethodGroup
    class NewArrayInit_MethodGroup:
        @typing.overload
        def __call__(self, type: typing.Type[typing.Any], initializers: Array_1[Expression]) -> NewArrayExpression:...
        @typing.overload
        def __call__(self, type: typing.Type[typing.Any], initializers: IEnumerable_1[Expression]) -> NewArrayExpression:...

    # Skipped Not due to it being static, abstract and generic.

    Not : Not_MethodGroup
    class Not_MethodGroup:
        @typing.overload
        def __call__(self, expression: Expression) -> UnaryExpression:...
        @typing.overload
        def __call__(self, expression: Expression, method: MethodInfo) -> UnaryExpression:...

    # Skipped NotEqual due to it being static, abstract and generic.

    NotEqual : NotEqual_MethodGroup
    class NotEqual_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, liftToNull: bool, method: MethodInfo) -> BinaryExpression:...

    # Skipped OnesComplement due to it being static, abstract and generic.

    OnesComplement : OnesComplement_MethodGroup
    class OnesComplement_MethodGroup:
        @typing.overload
        def __call__(self, expression: Expression) -> UnaryExpression:...
        @typing.overload
        def __call__(self, expression: Expression, method: MethodInfo) -> UnaryExpression:...

    # Skipped Or due to it being static, abstract and generic.

    Or : Or_MethodGroup
    class Or_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo) -> BinaryExpression:...

    # Skipped OrAssign due to it being static, abstract and generic.

    OrAssign : OrAssign_MethodGroup
    class OrAssign_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo, conversion: LambdaExpression) -> BinaryExpression:...

    # Skipped OrElse due to it being static, abstract and generic.

    OrElse : OrElse_MethodGroup
    class OrElse_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo) -> BinaryExpression:...

    # Skipped Parameter due to it being static, abstract and generic.

    Parameter : Parameter_MethodGroup
    class Parameter_MethodGroup:
        @typing.overload
        def __call__(self, type: typing.Type[typing.Any]) -> ParameterExpression:...
        @typing.overload
        def __call__(self, type: typing.Type[typing.Any], name: str) -> ParameterExpression:...

    # Skipped PostDecrementAssign due to it being static, abstract and generic.

    PostDecrementAssign : PostDecrementAssign_MethodGroup
    class PostDecrementAssign_MethodGroup:
        @typing.overload
        def __call__(self, expression: Expression) -> UnaryExpression:...
        @typing.overload
        def __call__(self, expression: Expression, method: MethodInfo) -> UnaryExpression:...

    # Skipped PostIncrementAssign due to it being static, abstract and generic.

    PostIncrementAssign : PostIncrementAssign_MethodGroup
    class PostIncrementAssign_MethodGroup:
        @typing.overload
        def __call__(self, expression: Expression) -> UnaryExpression:...
        @typing.overload
        def __call__(self, expression: Expression, method: MethodInfo) -> UnaryExpression:...

    # Skipped Power due to it being static, abstract and generic.

    Power : Power_MethodGroup
    class Power_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo) -> BinaryExpression:...

    # Skipped PowerAssign due to it being static, abstract and generic.

    PowerAssign : PowerAssign_MethodGroup
    class PowerAssign_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo, conversion: LambdaExpression) -> BinaryExpression:...

    # Skipped PreDecrementAssign due to it being static, abstract and generic.

    PreDecrementAssign : PreDecrementAssign_MethodGroup
    class PreDecrementAssign_MethodGroup:
        @typing.overload
        def __call__(self, expression: Expression) -> UnaryExpression:...
        @typing.overload
        def __call__(self, expression: Expression, method: MethodInfo) -> UnaryExpression:...

    # Skipped PreIncrementAssign due to it being static, abstract and generic.

    PreIncrementAssign : PreIncrementAssign_MethodGroup
    class PreIncrementAssign_MethodGroup:
        @typing.overload
        def __call__(self, expression: Expression) -> UnaryExpression:...
        @typing.overload
        def __call__(self, expression: Expression, method: MethodInfo) -> UnaryExpression:...

    # Skipped Property due to it being static, abstract and generic.

    Property : Property_MethodGroup
    class Property_MethodGroup:
        @typing.overload
        def __call__(self, expression: Expression, propertyAccessor: MethodInfo) -> MemberExpression:...
        @typing.overload
        def __call__(self, expression: Expression, property: PropertyInfo) -> MemberExpression:...
        @typing.overload
        def __call__(self, expression: Expression, propertyName: str) -> MemberExpression:...
        @typing.overload
        def __call__(self, instance: Expression, indexer: PropertyInfo, arguments: Array_1[Expression]) -> IndexExpression:...
        @typing.overload
        def __call__(self, instance: Expression, propertyName: str, arguments: Array_1[Expression]) -> IndexExpression:...
        @typing.overload
        def __call__(self, expression: Expression, type: typing.Type[typing.Any], propertyName: str) -> MemberExpression:...
        @typing.overload
        def __call__(self, instance: Expression, indexer: PropertyInfo, arguments: IEnumerable_1[Expression]) -> IndexExpression:...

    # Skipped Rethrow due to it being static, abstract and generic.

    Rethrow : Rethrow_MethodGroup
    class Rethrow_MethodGroup:
        @typing.overload
        def __call__(self) -> UnaryExpression:...
        @typing.overload
        def __call__(self, type: typing.Type[typing.Any]) -> UnaryExpression:...

    # Skipped Return due to it being static, abstract and generic.

    Return : Return_MethodGroup
    class Return_MethodGroup:
        @typing.overload
        def __call__(self, target: LabelTarget) -> GotoExpression:...
        @typing.overload
        def __call__(self, target: LabelTarget, type: typing.Type[typing.Any]) -> GotoExpression:...
        @typing.overload
        def __call__(self, target: LabelTarget, value: Expression) -> GotoExpression:...
        @typing.overload
        def __call__(self, target: LabelTarget, value: Expression, type: typing.Type[typing.Any]) -> GotoExpression:...

    # Skipped RightShift due to it being static, abstract and generic.

    RightShift : RightShift_MethodGroup
    class RightShift_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo) -> BinaryExpression:...

    # Skipped RightShiftAssign due to it being static, abstract and generic.

    RightShiftAssign : RightShiftAssign_MethodGroup
    class RightShiftAssign_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo, conversion: LambdaExpression) -> BinaryExpression:...

    # Skipped RuntimeVariables due to it being static, abstract and generic.

    RuntimeVariables : RuntimeVariables_MethodGroup
    class RuntimeVariables_MethodGroup:
        @typing.overload
        def __call__(self, variables: Array_1[ParameterExpression]) -> RuntimeVariablesExpression:...
        @typing.overload
        def __call__(self, variables: IEnumerable_1[ParameterExpression]) -> RuntimeVariablesExpression:...

    # Skipped Subtract due to it being static, abstract and generic.

    Subtract : Subtract_MethodGroup
    class Subtract_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo) -> BinaryExpression:...

    # Skipped SubtractAssign due to it being static, abstract and generic.

    SubtractAssign : SubtractAssign_MethodGroup
    class SubtractAssign_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo, conversion: LambdaExpression) -> BinaryExpression:...

    # Skipped SubtractAssignChecked due to it being static, abstract and generic.

    SubtractAssignChecked : SubtractAssignChecked_MethodGroup
    class SubtractAssignChecked_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo, conversion: LambdaExpression) -> BinaryExpression:...

    # Skipped SubtractChecked due to it being static, abstract and generic.

    SubtractChecked : SubtractChecked_MethodGroup
    class SubtractChecked_MethodGroup:
        @typing.overload
        def __call__(self, left: Expression, right: Expression) -> BinaryExpression:...
        @typing.overload
        def __call__(self, left: Expression, right: Expression, method: MethodInfo) -> BinaryExpression:...

    # Skipped Switch due to it being static, abstract and generic.

    Switch : Switch_MethodGroup
    class Switch_MethodGroup:
        @typing.overload
        def __call__(self, switchValue: Expression, cases: Array_1[SwitchCase]) -> SwitchExpression:...
        @typing.overload
        def __call__(self, switchValue: Expression, defaultBody: Expression, cases: Array_1[SwitchCase]) -> SwitchExpression:...
        @typing.overload
        def __call__(self, switchValue: Expression, defaultBody: Expression, comparison: MethodInfo, cases: Array_1[SwitchCase]) -> SwitchExpression:...
        @typing.overload
        def __call__(self, switchValue: Expression, defaultBody: Expression, comparison: MethodInfo, cases: IEnumerable_1[SwitchCase]) -> SwitchExpression:...
        @typing.overload
        def __call__(self, type: typing.Type[typing.Any], switchValue: Expression, defaultBody: Expression, comparison: MethodInfo, cases: Array_1[SwitchCase]) -> SwitchExpression:...
        @typing.overload
        def __call__(self, type: typing.Type[typing.Any], switchValue: Expression, defaultBody: Expression, comparison: MethodInfo, cases: IEnumerable_1[SwitchCase]) -> SwitchExpression:...

    # Skipped SwitchCase due to it being static, abstract and generic.

    SwitchCase : SwitchCase_MethodGroup
    class SwitchCase_MethodGroup:
        @typing.overload
        def __call__(self, body: Expression, testValues: Array_1[Expression]) -> SwitchCase:...
        @typing.overload
        def __call__(self, body: Expression, testValues: IEnumerable_1[Expression]) -> SwitchCase:...

    # Skipped SymbolDocument due to it being static, abstract and generic.

    SymbolDocument : SymbolDocument_MethodGroup
    class SymbolDocument_MethodGroup:
        @typing.overload
        def __call__(self, fileName: str) -> SymbolDocumentInfo:...
        @typing.overload
        def __call__(self, fileName: str, language: Guid) -> SymbolDocumentInfo:...
        @typing.overload
        def __call__(self, fileName: str, language: Guid, languageVendor: Guid) -> SymbolDocumentInfo:...
        @typing.overload
        def __call__(self, fileName: str, language: Guid, languageVendor: Guid, documentType: Guid) -> SymbolDocumentInfo:...

    # Skipped Throw due to it being static, abstract and generic.

    Throw : Throw_MethodGroup
    class Throw_MethodGroup:
        @typing.overload
        def __call__(self, value: Expression) -> UnaryExpression:...
        @typing.overload
        def __call__(self, value: Expression, type: typing.Type[typing.Any]) -> UnaryExpression:...

    # Skipped UnaryPlus due to it being static, abstract and generic.

    UnaryPlus : UnaryPlus_MethodGroup
    class UnaryPlus_MethodGroup:
        @typing.overload
        def __call__(self, expression: Expression) -> UnaryExpression:...
        @typing.overload
        def __call__(self, expression: Expression, method: MethodInfo) -> UnaryExpression:...

    # Skipped Variable due to it being static, abstract and generic.

    Variable : Variable_MethodGroup
    class Variable_MethodGroup:
        @typing.overload
        def __call__(self, type: typing.Type[typing.Any]) -> ParameterExpression:...
        @typing.overload
        def __call__(self, type: typing.Type[typing.Any], name: str) -> ParameterExpression:...



Expression_1_TDelegate = typing.TypeVar('Expression_1_TDelegate')
class Expression_1(typing.Generic[Expression_1_TDelegate], LambdaExpression):
    @property
    def Body(self) -> Expression: ...
    @property
    def CanReduce(self) -> bool: ...
    @property
    def Name(self) -> str: ...
    @property
    def NodeType(self) -> ExpressionType: ...
    @property
    def Parameters(self) -> ReadOnlyCollection_1[ParameterExpression]: ...
    @property
    def ReturnType(self) -> typing.Type[typing.Any]: ...
    @property
    def TailCall(self) -> bool: ...
    @property
    def Type(self) -> typing.Type[typing.Any]: ...
    def Update(self, body: Expression, parameters: IEnumerable_1[ParameterExpression]) -> Expression_1[Expression_1_TDelegate]: ...
    # Skipped Compile due to it being static, abstract and generic.

    Compile : Compile_MethodGroup[Expression_1_TDelegate]
    Compile_MethodGroup_Expression_1_TDelegate = typing.TypeVar('Compile_MethodGroup_Expression_1_TDelegate')
    class Compile_MethodGroup(typing.Generic[Compile_MethodGroup_Expression_1_TDelegate]):
        Compile_MethodGroup_Expression_1_TDelegate = Expression_1.Compile_MethodGroup_Expression_1_TDelegate
        @typing.overload
        def __call__(self) -> Compile_MethodGroup_Expression_1_TDelegate:...
        @typing.overload
        def __call__(self, preferInterpretation: bool) -> Compile_MethodGroup_Expression_1_TDelegate:...
        @typing.overload
        def __call__(self, debugInfoGenerator: DebugInfoGenerator) -> Compile_MethodGroup_Expression_1_TDelegate:...



class ExpressionType(typing.SupportsInt):
    @typing.overload
    def __init__(self, value : int) -> None: ...
    @typing.overload
    def __init__(self, value : int, force_if_true: bool) -> None: ...
    def __int__(self) -> int: ...
    
    # Values:
    Add : ExpressionType # 0
    AddChecked : ExpressionType # 1
    And : ExpressionType # 2
    AndAlso : ExpressionType # 3
    ArrayLength : ExpressionType # 4
    ArrayIndex : ExpressionType # 5
    Call : ExpressionType # 6
    Coalesce : ExpressionType # 7
    Conditional : ExpressionType # 8
    Constant : ExpressionType # 9
    Convert : ExpressionType # 10
    ConvertChecked : ExpressionType # 11
    Divide : ExpressionType # 12
    Equal : ExpressionType # 13
    ExclusiveOr : ExpressionType # 14
    GreaterThan : ExpressionType # 15
    GreaterThanOrEqual : ExpressionType # 16
    Invoke : ExpressionType # 17
    Lambda : ExpressionType # 18
    LeftShift : ExpressionType # 19
    LessThan : ExpressionType # 20
    LessThanOrEqual : ExpressionType # 21
    ListInit : ExpressionType # 22
    MemberAccess : ExpressionType # 23
    MemberInit : ExpressionType # 24
    Modulo : ExpressionType # 25
    Multiply : ExpressionType # 26
    MultiplyChecked : ExpressionType # 27
    Negate : ExpressionType # 28
    UnaryPlus : ExpressionType # 29
    NegateChecked : ExpressionType # 30
    New : ExpressionType # 31
    NewArrayInit : ExpressionType # 32
    NewArrayBounds : ExpressionType # 33
    Not : ExpressionType # 34
    NotEqual : ExpressionType # 35
    Or : ExpressionType # 36
    OrElse : ExpressionType # 37
    Parameter : ExpressionType # 38
    Power : ExpressionType # 39
    Quote : ExpressionType # 40
    RightShift : ExpressionType # 41
    Subtract : ExpressionType # 42
    SubtractChecked : ExpressionType # 43
    TypeAs : ExpressionType # 44
    TypeIs : ExpressionType # 45
    Assign : ExpressionType # 46
    Block : ExpressionType # 47
    DebugInfo : ExpressionType # 48
    Decrement : ExpressionType # 49
    Dynamic : ExpressionType # 50
    Default : ExpressionType # 51
    Extension : ExpressionType # 52
    Goto : ExpressionType # 53
    Increment : ExpressionType # 54
    Index : ExpressionType # 55
    Label : ExpressionType # 56
    RuntimeVariables : ExpressionType # 57
    Loop : ExpressionType # 58
    Switch : ExpressionType # 59
    Throw : ExpressionType # 60
    Try : ExpressionType # 61
    Unbox : ExpressionType # 62
    AddAssign : ExpressionType # 63
    AndAssign : ExpressionType # 64
    DivideAssign : ExpressionType # 65
    ExclusiveOrAssign : ExpressionType # 66
    LeftShiftAssign : ExpressionType # 67
    ModuloAssign : ExpressionType # 68
    MultiplyAssign : ExpressionType # 69
    OrAssign : ExpressionType # 70
    PowerAssign : ExpressionType # 71
    RightShiftAssign : ExpressionType # 72
    SubtractAssign : ExpressionType # 73
    AddAssignChecked : ExpressionType # 74
    MultiplyAssignChecked : ExpressionType # 75
    SubtractAssignChecked : ExpressionType # 76
    PreIncrementAssign : ExpressionType # 77
    PreDecrementAssign : ExpressionType # 78
    PostIncrementAssign : ExpressionType # 79
    PostDecrementAssign : ExpressionType # 80
    TypeEqual : ExpressionType # 81
    OnesComplement : ExpressionType # 82
    IsTrue : ExpressionType # 83
    IsFalse : ExpressionType # 84


class GotoExpression(Expression_0):
    @property
    def CanReduce(self) -> bool: ...
    @property
    def Kind(self) -> GotoExpressionKind: ...
    @property
    def NodeType(self) -> ExpressionType: ...
    @property
    def Target(self) -> LabelTarget: ...
    @property
    def Type(self) -> typing.Type[typing.Any]: ...
    @property
    def Value(self) -> Expression: ...
    def Update(self, target: LabelTarget, value: Expression) -> GotoExpression: ...


class GotoExpressionKind(typing.SupportsInt):
    @typing.overload
    def __init__(self, value : int) -> None: ...
    @typing.overload
    def __init__(self, value : int, force_if_true: bool) -> None: ...
    def __int__(self) -> int: ...
    
    # Values:
    Goto : GotoExpressionKind # 0
    Return : GotoExpressionKind # 1
    Break : GotoExpressionKind # 2
    Continue : GotoExpressionKind # 3


class IArgumentProvider(typing.Protocol):
    @property
    def ArgumentCount(self) -> int: ...
    @abc.abstractmethod
    def GetArgument(self, index: int) -> Expression: ...


class IDynamicExpression(IArgumentProvider, typing.Protocol):
    @property
    def DelegateType(self) -> typing.Type[typing.Any]: ...
    @abc.abstractmethod
    def CreateCallSite(self) -> typing.Any: ...
    @abc.abstractmethod
    def Rewrite(self, args: Array_1[Expression]) -> Expression: ...


class IndexExpression(Expression_0, IArgumentProvider):
    @property
    def ArgumentCount(self) -> int: ...
    @property
    def Arguments(self) -> ReadOnlyCollection_1[Expression]: ...
    @property
    def CanReduce(self) -> bool: ...
    @property
    def Indexer(self) -> PropertyInfo: ...
    @property
    def NodeType(self) -> ExpressionType: ...
    @property
    def Object(self) -> Expression: ...
    @property
    def Type(self) -> typing.Type[typing.Any]: ...
    def GetArgument(self, index: int) -> Expression: ...
    def Update(self, object: Expression, arguments: IEnumerable_1[Expression]) -> IndexExpression: ...


class InvocationExpression(Expression_0, IArgumentProvider):
    @property
    def ArgumentCount(self) -> int: ...
    @property
    def Arguments(self) -> ReadOnlyCollection_1[Expression]: ...
    @property
    def CanReduce(self) -> bool: ...
    @property
    def Expression(self) -> Expression: ...
    @property
    def NodeType(self) -> ExpressionType: ...
    @property
    def Type(self) -> typing.Type[typing.Any]: ...
    def GetArgument(self, index: int) -> Expression: ...
    def Update(self, expression: Expression, arguments: IEnumerable_1[Expression]) -> InvocationExpression: ...


class LabelExpression(Expression_0):
    @property
    def CanReduce(self) -> bool: ...
    @property
    def DefaultValue(self) -> Expression: ...
    @property
    def NodeType(self) -> ExpressionType: ...
    @property
    def Target(self) -> LabelTarget: ...
    @property
    def Type(self) -> typing.Type[typing.Any]: ...
    def Update(self, target: LabelTarget, defaultValue: Expression) -> LabelExpression: ...


class LabelTarget:
    @property
    def Name(self) -> str: ...
    @property
    def Type(self) -> typing.Type[typing.Any]: ...
    def ToString(self) -> str: ...


class LambdaExpression(Expression_0):
    @property
    def Body(self) -> Expression: ...
    @classmethod
    @property
    def CanCompileToIL(cls) -> bool: ...
    @classmethod
    @property
    def CanInterpret(cls) -> bool: ...
    @property
    def CanReduce(self) -> bool: ...
    @property
    def Name(self) -> str: ...
    @property
    def NodeType(self) -> ExpressionType: ...
    @property
    def Parameters(self) -> ReadOnlyCollection_1[ParameterExpression]: ...
    @property
    def ReturnType(self) -> typing.Type[typing.Any]: ...
    @property
    def TailCall(self) -> bool: ...
    @property
    def Type(self) -> typing.Type[typing.Any]: ...
    # Skipped Compile due to it being static, abstract and generic.

    Compile : Compile_MethodGroup
    class Compile_MethodGroup:
        @typing.overload
        def __call__(self) -> Delegate:...
        @typing.overload
        def __call__(self, preferInterpretation: bool) -> Delegate:...
        @typing.overload
        def __call__(self, debugInfoGenerator: DebugInfoGenerator) -> Delegate:...



class ListInitExpression(Expression_0):
    @property
    def CanReduce(self) -> bool: ...
    @property
    def Initializers(self) -> ReadOnlyCollection_1[ElementInit]: ...
    @property
    def NewExpression(self) -> NewExpression: ...
    @property
    def NodeType(self) -> ExpressionType: ...
    @property
    def Type(self) -> typing.Type[typing.Any]: ...
    def Reduce(self) -> Expression: ...
    def Update(self, newExpression: NewExpression, initializers: IEnumerable_1[ElementInit]) -> ListInitExpression: ...


class LoopExpression(Expression_0):
    @property
    def Body(self) -> Expression: ...
    @property
    def BreakLabel(self) -> LabelTarget: ...
    @property
    def CanReduce(self) -> bool: ...
    @property
    def ContinueLabel(self) -> LabelTarget: ...
    @property
    def NodeType(self) -> ExpressionType: ...
    @property
    def Type(self) -> typing.Type[typing.Any]: ...
    def Update(self, breakLabel: LabelTarget, continueLabel: LabelTarget, body: Expression) -> LoopExpression: ...


class MemberAssignment(MemberBinding):
    @property
    def BindingType(self) -> MemberBindingType: ...
    @property
    def Expression(self) -> Expression: ...
    @property
    def Member(self) -> MemberInfo: ...
    def Update(self, expression: Expression) -> MemberAssignment: ...


class MemberBinding(abc.ABC):
    @property
    def BindingType(self) -> MemberBindingType: ...
    @property
    def Member(self) -> MemberInfo: ...
    def ToString(self) -> str: ...


class MemberBindingType(typing.SupportsInt):
    @typing.overload
    def __init__(self, value : int) -> None: ...
    @typing.overload
    def __init__(self, value : int, force_if_true: bool) -> None: ...
    def __int__(self) -> int: ...
    
    # Values:
    Assignment : MemberBindingType # 0
    MemberBinding : MemberBindingType # 1
    ListBinding : MemberBindingType # 2


class MemberExpression(Expression_0):
    @property
    def CanReduce(self) -> bool: ...
    @property
    def Expression(self) -> Expression: ...
    @property
    def Member(self) -> MemberInfo: ...
    @property
    def NodeType(self) -> ExpressionType: ...
    @property
    def Type(self) -> typing.Type[typing.Any]: ...
    def Update(self, expression: Expression) -> MemberExpression: ...


class MemberInitExpression(Expression_0):
    @property
    def Bindings(self) -> ReadOnlyCollection_1[MemberBinding]: ...
    @property
    def CanReduce(self) -> bool: ...
    @property
    def NewExpression(self) -> NewExpression: ...
    @property
    def NodeType(self) -> ExpressionType: ...
    @property
    def Type(self) -> typing.Type[typing.Any]: ...
    def Reduce(self) -> Expression: ...
    def Update(self, newExpression: NewExpression, bindings: IEnumerable_1[MemberBinding]) -> MemberInitExpression: ...


class MemberListBinding(MemberBinding):
    @property
    def BindingType(self) -> MemberBindingType: ...
    @property
    def Initializers(self) -> ReadOnlyCollection_1[ElementInit]: ...
    @property
    def Member(self) -> MemberInfo: ...
    def Update(self, initializers: IEnumerable_1[ElementInit]) -> MemberListBinding: ...


class MemberMemberBinding(MemberBinding):
    @property
    def Bindings(self) -> ReadOnlyCollection_1[MemberBinding]: ...
    @property
    def BindingType(self) -> MemberBindingType: ...
    @property
    def Member(self) -> MemberInfo: ...
    def Update(self, bindings: IEnumerable_1[MemberBinding]) -> MemberMemberBinding: ...


class MethodCallExpression(Expression_0, IArgumentProvider):
    @property
    def ArgumentCount(self) -> int: ...
    @property
    def Arguments(self) -> ReadOnlyCollection_1[Expression]: ...
    @property
    def CanReduce(self) -> bool: ...
    @property
    def Method(self) -> MethodInfo: ...
    @property
    def NodeType(self) -> ExpressionType: ...
    @property
    def Object(self) -> Expression: ...
    @property
    def Type(self) -> typing.Type[typing.Any]: ...
    def GetArgument(self, index: int) -> Expression: ...
    def Update(self, object: Expression, arguments: IEnumerable_1[Expression]) -> MethodCallExpression: ...


class NewArrayExpression(Expression_0):
    @property
    def CanReduce(self) -> bool: ...
    @property
    def Expressions(self) -> ReadOnlyCollection_1[Expression]: ...
    @property
    def NodeType(self) -> ExpressionType: ...
    @property
    def Type(self) -> typing.Type[typing.Any]: ...
    def Update(self, expressions: IEnumerable_1[Expression]) -> NewArrayExpression: ...


class NewExpression(Expression_0, IArgumentProvider):
    @property
    def ArgumentCount(self) -> int: ...
    @property
    def Arguments(self) -> ReadOnlyCollection_1[Expression]: ...
    @property
    def CanReduce(self) -> bool: ...
    @property
    def Constructor(self) -> ConstructorInfo: ...
    @property
    def Members(self) -> ReadOnlyCollection_1[MemberInfo]: ...
    @property
    def NodeType(self) -> ExpressionType: ...
    @property
    def Type(self) -> typing.Type[typing.Any]: ...
    def GetArgument(self, index: int) -> Expression: ...
    def Update(self, arguments: IEnumerable_1[Expression]) -> NewExpression: ...


class ParameterExpression(Expression_0):
    @property
    def CanReduce(self) -> bool: ...
    @property
    def IsByRef(self) -> bool: ...
    @property
    def Name(self) -> str: ...
    @property
    def NodeType(self) -> ExpressionType: ...
    @property
    def Type(self) -> typing.Type[typing.Any]: ...


class RuntimeVariablesExpression(Expression_0):
    @property
    def CanReduce(self) -> bool: ...
    @property
    def NodeType(self) -> ExpressionType: ...
    @property
    def Type(self) -> typing.Type[typing.Any]: ...
    @property
    def Variables(self) -> ReadOnlyCollection_1[ParameterExpression]: ...
    def Update(self, variables: IEnumerable_1[ParameterExpression]) -> RuntimeVariablesExpression: ...


class SwitchCase:
    @property
    def Body(self) -> Expression: ...
    @property
    def TestValues(self) -> ReadOnlyCollection_1[Expression]: ...
    def ToString(self) -> str: ...
    def Update(self, testValues: IEnumerable_1[Expression], body: Expression) -> SwitchCase: ...


class SwitchExpression(Expression_0):
    @property
    def CanReduce(self) -> bool: ...
    @property
    def Cases(self) -> ReadOnlyCollection_1[SwitchCase]: ...
    @property
    def Comparison(self) -> MethodInfo: ...
    @property
    def DefaultBody(self) -> Expression: ...
    @property
    def NodeType(self) -> ExpressionType: ...
    @property
    def SwitchValue(self) -> Expression: ...
    @property
    def Type(self) -> typing.Type[typing.Any]: ...
    def Update(self, switchValue: Expression, cases: IEnumerable_1[SwitchCase], defaultBody: Expression) -> SwitchExpression: ...


class SymbolDocumentInfo:
    @property
    def DocumentType(self) -> Guid: ...
    @property
    def FileName(self) -> str: ...
    @property
    def Language(self) -> Guid: ...
    @property
    def LanguageVendor(self) -> Guid: ...


class TryExpression(Expression_0):
    @property
    def Body(self) -> Expression: ...
    @property
    def CanReduce(self) -> bool: ...
    @property
    def Fault(self) -> Expression: ...
    @property
    def Finally(self) -> Expression: ...
    @property
    def Handlers(self) -> ReadOnlyCollection_1[CatchBlock]: ...
    @property
    def NodeType(self) -> ExpressionType: ...
    @property
    def Type(self) -> typing.Type[typing.Any]: ...
    def Update(self, body: Expression, handlers: IEnumerable_1[CatchBlock], finally: Expression, fault: Expression) -> TryExpression: ...


class TypeBinaryExpression(Expression_0):
    @property
    def CanReduce(self) -> bool: ...
    @property
    def Expression(self) -> Expression: ...
    @property
    def NodeType(self) -> ExpressionType: ...
    @property
    def Type(self) -> typing.Type[typing.Any]: ...
    @property
    def TypeOperand(self) -> typing.Type[typing.Any]: ...
    def Update(self, expression: Expression) -> TypeBinaryExpression: ...


class UnaryExpression(Expression_0):
    @property
    def CanReduce(self) -> bool: ...
    @property
    def IsLifted(self) -> bool: ...
    @property
    def IsLiftedToNull(self) -> bool: ...
    @property
    def Method(self) -> MethodInfo: ...
    @property
    def NodeType(self) -> ExpressionType: ...
    @property
    def Operand(self) -> Expression: ...
    @property
    def Type(self) -> typing.Type[typing.Any]: ...
    def Reduce(self) -> Expression: ...
    def Update(self, operand: Expression) -> UnaryExpression: ...

